// copyright 2002 Josh Levenberg, please see copying.txt

/* In order to cache on a NVidia card we need to do a few things:
   - allocate video memory: unfortunately, we can't re-allocate or use more
    than one block and still get good performance.  Therefore we grab as
    big a block as we can.
   - put vertex data in the allocated video memory: I divide it into 
   maxnumvert*FloatsPerVert*sizeof(float) chunks.  This avoids fragmentation
   entirely.
   - wait for the video card to be done with a chunk before re-using it: 
   this is done using fences: we set a fence in the command stream and later
   query it.  If the video card has rendered up to that point, we get a true
   return and we can re-use any memory that is not used in any later commmand.
   The Discard command won't typically mark video memory for re-use 
   immediately, we have to wait for the fence after its last use.

   More on that last point:
   1) we start out by creating a "FenceGroup"
   2) as we render caches, we set their "LastRender" to the current FenceGroup
   3) after we have rendered VertsPerGroup vertices, we set a fence and store
     it in the FenceGroup.  
   4) we create a new FenceGroup.  This will now be the current FenceGroup.
   5) we render caches as before, but now their "LastRender" is set to the new
     FenceGroup
   6) if a cache is discarded, we check its "LastRender" value
      a) if it is less than mFirstFenceGroup, we can put this cache on the 
        free (ready to be reused) list immediately
      b) otherwise it is added to the linked list of "Pending" caches in the
        FenceGroup it was last rendered in
   7) after the second FenceGroup is full, we set a fence as before.  We also
     check the first FenceGroup to see if the fence has been passed.  If it 
     has we:
      a) move its pending list to the free list
      b) discard its fence
      c) increment mFirstFenceGroup
      d) check the new first FenceGroup to see if the fence has passed, if
        so, go back to (a)
   8) Go back to step 4
 */

#include "disp_nv.h"
#include "dlod.h"
#include "model.h"
#include <assert.h>
//#include <stdio.h>
#include <stdlib.h>
#include "glext.h"

enum { VertsPerGroup = (1<<17) };
enum { MaxBytesVideoRam = (3<<24) };

#ifndef NDEBUG
static int sNumBegin=0;
static int sNumFinish=0;
static int sNumDiscard=0;
#endif

NvidiaDisplay::NvidiaDisplay()
  : mCurrentCache(-1), mVideoMemory(0)
  {
  }

void NvidiaDisplay::Initialize(int maxnumvert, int maxnumind)
  { 
    assert(mVideoMemory==0);
    assert(mCurrentCache==-1);
    assert(mCaches.empty());
    assert(mFenceGroups.empty());

    mBlockSize=maxnumvert;
    mDefaultTriMemory=maxnumind;
    if (mBlockSize>MaxIndex)
      {
        printf("Block size greater than maximum index range\n");
        exit(2);
      }

    mFirstFenceGroup=0;
    mFirstFreeCache=-1; 
    mFenceGroups.push_back();
    mFenceGroups.back().mFirstPending = -1;
    mFenceGroups.back().mLastPending = -1;

    mCachesPerGroup = VertsPerGroup / maxnumvert;
    if (mCachesPerGroup < 1)
      mCachesPerGroup = 1;
    mNumLeftLastGroup=mCachesPerGroup;

    // figure out how much video memory to allocate
    // [set mNumBlocks, mVideoMemory]
    mVideoMemory=0;
    bool videomem=true;
    unsigned sizeinbytes=0;
    do
      {
        mNumBlocks=1;
            // priority .5 -> AGP, 1 -> video
        float priority = videomem ? 1 : .5;
        do
          {
            if (mVideoMemory)
              wglFreeMemoryNV(mVideoMemory);
            sizeinbytes = sizeof(float)*FloatsPerVert*mBlockSize*mNumBlocks;
            mVideoMemory = static_cast<float *>(wglAllocateMemoryNV(sizeinbytes, 0, 0, priority));
            mNumBlocks <<=1;
          } while (mVideoMemory && sizeinbytes<= MaxBytesVideoRam);
        if (mVideoMemory)
          mNumBlocks >>=1;
        else
          {
            sizeinbytes >>= 1;
            mNumBlocks >>=2;
            if (mNumBlocks)
              mVideoMemory = static_cast<float *>(wglAllocateMemoryNV(sizeinbytes, 0, 0, priority));
            if (!mVideoMemory)
              if (videomem)
                {
                  videomem=false;
                  fprintf(gOutFile, "Could not allocate video memory, switching to AGP memory.\n");
                }
              else
                {
                  fprintf(gOutFile, "Failure to allocate cache memory.\n");
                  exit(2);
                }
          }
      }  while (! mVideoMemory);

    assert(sizeinbytes);
    fprintf(gOutFile, "Successfully got %d bytes (%dk) of video memory\n", 
            sizeinbytes, sizeinbytes/1024);

    glVertexArrayRangeNV(sizeinbytes, mVideoMemory);

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY_RANGE_NV);

    // set VertexPointer, TextureCoordPointer
    SetVidMemOffsetPtr(mVideoMemory);

#ifndef NDEBUG
    sNumBegin=0;
    sNumFinish=0;
    sNumDiscard=0;
    CheckBlockConsistency();
#endif
  }

void NvidiaDisplay::Done()
  { 
    assert(mCurrentCache==-1);
    glFinish();
    fprintf(gOutFile, "Actually used %dk video memory\n", sizeof(float)*FloatsPerVert*mBlockSize*mCaches.size()/1024);
    for (int i=0; i<mCaches.size(); ++i)
      delete[] mCaches[i].mTrisBegin;
    mCaches.clear();

    for (i=0; i<mFenceGroups.size()-1; ++i)
      glDeleteFencesNV(1, & mFenceGroups[i].mFence);
    mFenceGroups.clear();

    glVertexArrayRangeNV(0, 0);
    wglFreeMemoryNV(mVideoMemory);
    glDisableClientState(GL_VERTEX_ARRAY_RANGE_NV);
    mVideoMemory=0;
  }

void NvidiaDisplay::PopGroup()
  {
    glDeleteFencesNV(1, & mFenceGroups.front().mFence);
    assert((mFenceGroups.front().mLastPending==-1) ==
           (mFenceGroups.front().mFirstPending==-1));
    int lastpending;
    if ((lastpending=mFenceGroups.front().mLastPending)!=-1)
      {
        assert(mCaches[lastpending].u.Next==-1);
        mCaches[lastpending].u.Next=mFirstFreeCache;
        mFirstFreeCache=mFenceGroups.front().mFirstPending;
      }
    mFenceGroups.pop_front();
    ++mFirstFenceGroup;
  }

bool NvidiaDisplay::BeginCache(Coord *&verts, Index *&inds)
  {
    assert(mCurrentCache==-1);
#ifndef NDEBUG
    CheckBlockConsistency();
#endif

    while (mFirstFreeCache == -1 && mFenceGroups.size()>1 && 
             glTestFenceNV(mFenceGroups.front().mFence))
      {
        PopGroup();
      }
    if (mFirstFreeCache != -1)
      { // use an existing cache block that is free
        mCurrentCache=mFirstFreeCache;
        mFirstFreeCache=mCaches[mFirstFreeCache].u.Next;
      }
    else if (mCaches.size() < mNumBlocks)
      { // allocate a new cache block
        mCurrentCache=mCaches.size();
        mCaches.push_back();
        mCaches.back().mTrisBegin=new indextype[mDefaultTriMemory];
      }
    else 
      {
        while (mFenceGroups.size()>1)
          {
            glFinishFenceNV(mFenceGroups.front().mFence);
            PopGroup();
            if (mFirstFreeCache!=-1)
              {
                mCurrentCache=mFirstFreeCache;
                mFirstFreeCache=mCaches[mFirstFreeCache].u.Next;
                break;
              }
          }
        if (mCurrentCache==-1)
          {
            CloseLastGroup();
            glFinishFenceNV(mFenceGroups.front().mFence);
            PopGroup();
            if (mFirstFreeCache!=-1)
              {
                mCurrentCache=mFirstFreeCache;
                mFirstFreeCache=mCaches[mFirstFreeCache].u.Next;
              }
            else
              {
                fprintf(gOutFile, "Ran out of video memory.\n");
                exit(1);
                return false;
              }
          }
      }
    // mCurrentCache is set
    // mCaches[mCurrentCache].mTrisBegin is set

    // return memory pools
    verts=mVideoMemory+mCurrentCache*mBlockSize*FloatsPerVert;
    inds=mCaches[mCurrentCache].mTrisBegin;
#ifndef NDEBUG
    mCaches[mCurrentCache].mDiscarded=false;
#endif

    // indicies are >=0 && < 65536
    SetVidMemOffsetPtr(verts);

#ifndef NDEBUG
    ++sNumBegin;
    CheckBlockConsistency();
#endif
    return true;
  }

Cache NvidiaDisplay::FinishCache(int numvert, int numind)
  {
    assert(mCurrentCache>=0 && mCurrentCache<mNumBlocks && mCurrentCache<mCaches.size());
    assert(numvert>=0);
    assert(numvert<=mBlockSize);
    assert(numind>=0);
    assert(numind<=mDefaultTriMemory);
#ifndef NDEBUG
    ++sNumFinish;
    for (int i=0; i<numind; ++i)
      {
        int ind=mCaches[mCurrentCache].mTrisBegin[i];
        assert(ind>=0);
        assert(ind<numvert);
      }
#endif
    // anything else?
    mCaches[mCurrentCache].mTrisEnd=mCaches[mCurrentCache].mTrisBegin+numind;

#ifdef RANGE_ELEMENTS_NV
    mCaches[mCurrentCache].mNumVerts=numvert;
    glDrawRangeElements(GL_TRIANGLES, 0, numvert, numind, 
                        indexenum, mCaches[mCurrentCache].mTrisBegin);
#else
    glDrawElements(GL_TRIANGLES, numind, indexenum, 
                    mCaches[mCurrentCache].mTrisBegin);
#endif
    Cache ret=mCurrentCache;
    mCurrentCache= -1;

    AddToLastGroup(ret);

    return ret;
  }

void NvidiaDisplay::CloseLastGroup()
  {
    GLuint fence;
    glGenFencesNV(1, & fence);
    glSetFenceNV(fence, GL_ALL_COMPLETED_NV);
    mFenceGroups.back().mFence=fence;
    // test first group to see if it is ready to pop
    while (mFenceGroups.size()>1 && glTestFenceNV(mFenceGroups.front().mFence))
      {
        PopGroup();
      }
    // open new last group
    mFenceGroups.push_back();
    mFenceGroups.back().mFirstPending=-1;
    mFenceGroups.back().mLastPending=-1;
    mNumLeftLastGroup=mCachesPerGroup;
  }

void NvidiaDisplay::AddToLastGroup(int cache)
  {
    assert(mCurrentCache==-1);
    assert(mCaches[cache].mDiscarded==false);
    mCaches[cache].u.LastRender=mFirstFenceGroup+mFenceGroups.size()-1;
    if (! --mNumLeftLastGroup)
      CloseLastGroup();    
  }


void NvidiaDisplay::RenderCache(Cache cache)
  {
    assert(mCurrentCache==-1);
#ifndef NDEBUG
    assert(mCaches[cache].mDiscarded==false);
    CheckBlockConsistency();
#endif
    assert(cache>=0 && cache<mNumBlocks && cache<mCaches.size());
    SetVidMemOffsetPtr(mVideoMemory+cache*mBlockSize*FloatsPerVert);
    indextype *ptr = mCaches[cache].mTrisBegin;
    indextype *end = mCaches[cache].mTrisEnd;

#ifdef RANGE_ELEMENTS_NV
    glDrawRangeElements(GL_TRIANGLES, 0 ,
                        mCaches[cache].mMaxInd, 
                        end-ptr, indexenum, ptr);
#else
    glDrawElements(GL_TRIANGLES, end-ptr, indexenum, ptr);
#endif

    AddToLastGroup(cache);
#ifndef NDEBUG
    CheckBlockConsistency();
#endif
  }

void NvidiaDisplay::DiscardCache(Cache cache)
  {
    assert(mCurrentCache==-1);
#ifndef NDEBUG
    CheckBlockConsistency();
#endif
    assert(cache>=0 && cache<mNumBlocks && cache<mCaches.size());
    int lastrender=mCaches[cache].u.LastRender - mFirstFenceGroup;
    if (lastrender<0)
      { // already not in use, add to free list
        mCaches[cache].u.Next=mFirstFreeCache;
        mFirstFreeCache=cache;
      }
    else 
      {
        assert(lastrender<mFenceGroups.size());
        int firstpending=mFenceGroups[lastrender].mFirstPending;
        mCaches[cache].u.Next=firstpending;
        mFenceGroups[lastrender].mFirstPending=cache;
        if (firstpending==-1)
          {
            assert(mFenceGroups[lastrender].mLastPending==-1);
            mFenceGroups[lastrender].mLastPending=cache;
          }
      }

#ifndef NDEBUG
    ++sNumDiscard;
    mCaches[cache].mDiscarded=true;
    CheckBlockConsistency();
#endif
  }

#ifndef NDEBUG
void NvidiaDisplay::CheckBlockConsistency()
  {
    int discarded=0;
    if (mFirstFreeCache!=-1)
      {
        // iterate through recently freed blocks
        for (int cache = mFirstFreeCache;
             cache != -1;
             cache = mCaches[cache].u.Next)
          {
            assert(cache>=0);
            assert(cache<mCaches.size());
            assert(cache<mNumBlocks);
            ++discarded;
          }
      }
    for (int i=0; i<mFenceGroups.size(); ++i)
      {
        for (int cache = mFenceGroups[i].mFirstPending;
            cache != -1;
            cache = mCaches[cache].u.Next)
          {
            assert(cache>=0);
            assert(cache<mCaches.size());
            assert(cache<mNumBlocks);
            ++discarded;
          }
      }
    int cs=mCaches.size();
    assert(sNumBegin-sNumDiscard == mCaches.size()-discarded);
    int discarded2=0;
    for (int cache=0; cache < cs; ++cache)
      if (mCaches[cache].mDiscarded)
        ++discarded2;
    assert(discarded==discarded2);
  }
#endif

