// copyright 2002 Josh Levenberg, please see copying.txt
#include <assert.h>

/* Templated version:
template <int ShiftT>
class WrappingHeightField
  {
public:
    enum { ShiftAmt=ShiftT, Size=(1<<ShiftAmt), Mask=(Size-1) } ;

    typedef int heighttype;

protected:
    heighttype mHeights[Size*Size];

public:
    typedef const heighttype *const_iterator;
    typedef const_iterator iterator;
    iterator Begin() const { return mHeights; }
    iterator End()   const { return Begin()+Size*Size; }
    heighttype operator()(int x, int y)
      { return mHeights[(x&Mask)+((y&Mask)<<ShiftAmt)]; }
    void Set(int x, int y, heighttype h)
      { assert(((unsigned)x)<Size);
        assert(((unsigned)y)<Size);
        mHeights[x+(y<<ShiftAmt)]=h;
      }

    void SqSqFractal(heighttype roughness);
  };

template <int ShiftT>
void WrappingHeightField :: SqSqFractal(heighttype roughness)
*/

class WrappingHeightField
  {
public:
    typedef int heighttype;

protected:
    int mShiftAmt, mSize, mMask, mSquareSize;
    heighttype *mHeights;

    void copy(const WrappingHeightField &cpy);

public:
    explicit WrappingHeightField(int shiftAmt) 
          : mShiftAmt(shiftAmt), 
            mSize(1<<mShiftAmt), 
            mMask(mSize-1), 
            mSquareSize(mSize*mSize)
      { assert(shiftAmt>=1); mHeights=new heighttype[mSquareSize]; }
    WrappingHeightField(const WrappingHeightField &cpy)
      { copy(cpy); }
    void operator=(const WrappingHeightField &cpy)
      { delete[] mHeights; copy(cpy); }
    ~WrappingHeightField()
      { delete[] mHeights; }

    typedef const heighttype *const_iterator;
    typedef const_iterator iterator;
    iterator Begin() const { return mHeights; }
    iterator End()   const { return Begin()+mSquareSize; }

    int Size() const { return mSize; }

    heighttype operator()(int x, int y) const
      { return mHeights[(x&mMask)+((y&mMask)<<mShiftAmt)]; }
    void Set(int x, int y, heighttype h)
      { assert(((unsigned)x)<(unsigned)mSize);
        assert(((unsigned)y)<(unsigned)mSize);
        mHeights[x+(y<<mShiftAmt)]=h;
      }

    void GetMinMax(heighttype &minH, heighttype &maxH) const;
    void Normalize(heighttype minH, heighttype maxH);
    void Exp(heighttype minH, heighttype maxH);
    void Square();

    static void SeedRandomizer(int seed);
    void SqSqFractal(heighttype noisescale, double noisefactor=2.0);
    void MultiFractal(const WrappingHeightField &roughness);

    void ScaleDown(const WrappingHeightField &original);
    void AddScaled(const WrappingHeightField &toadd, float weight);

      // for example 1<<24, 1.414, 1<<24, 1<<30
      // roughnessvaryscale controls how many mountains & valleys there are
      // roughnessvaryfactor is relatively unimportant
      // minroughness & maxroughness control how random/smooth the terrain is
    void JFractal(heighttype roughnessvaryscale, double roughnessvaryfactor, 
                  heighttype minroughness, heighttype maxroughness);
    void ColorJFractal(heighttype noisevaryscale, double noisevaryfactor, 
                                     heighttype minroughness, heighttype maxroughness,
                                     WrappingHeightField &color);


#if 0
    void SetToSmoothed(const WrappingHeightField &hf, double strength=1.0);
#else
    void SetToSmoothed(const WrappingHeightField &hf, int width=-1);
#endif

protected:
    
    static heighttype Random(int max);
  };
