// copyright 2002 Josh Levenberg, please see copying.txt
#include "fractal.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>

void WrappingHeightField :: SeedRandomizer(int seed)
  {
    srand(seed);
  }

WrappingHeightField::heighttype WrappingHeightField :: Random(int max)
  {
    int range=(max<<1)+1;
    /* generate a random number from 0 to range-1 */
    double random=((double)rand())/(RAND_MAX+1);
    return ((int) (random*range-max));
  }

void WrappingHeightField :: copy(const WrappingHeightField &cpy)
  {
    mShiftAmt=(cpy.mShiftAmt);
    assert(mShiftAmt>=1); 
    mSize=(1<<mShiftAmt);
    mMask=(mSize-1);
    mSquareSize=(mSize*mSize);
    mHeights=new heighttype[mSquareSize]; 
    memcpy(mHeights, cpy.mHeights, sizeof(heighttype)*mSquareSize);
 }

void WrappingHeightField :: GetMinMax(heighttype &minH, heighttype &maxH) const
  {
    iterator cur=Begin();
    maxH=minH=*cur;
    for (++cur; cur!=End(); ++cur)
      if (*cur>maxH)
        maxH=*cur;
      else if (*cur<minH)
        minH=*cur;
  }

void WrappingHeightField :: Normalize(heighttype minH, heighttype maxH)
  {
    heighttype origMin, origMax;
    GetMinMax(origMin, origMax);
    double scalef=((double) maxH-minH+1)/(origMax-origMin+1);

    for (heighttype *cur=mHeights; cur!=mHeights+mSquareSize; ++cur)
      *cur=(heighttype)((*cur-origMin)*scalef+minH);
  }

void WrappingHeightField :: Exp(heighttype minH, heighttype maxH)
  {
    heighttype origMin, origMax;
    GetMinMax(origMin, origMax);

    double scalef=(log(maxH)-log(minH))/(origMax-origMin);
    double toadd=log(minH);
    for (heighttype *cur=mHeights; cur!=mHeights+mSquareSize; ++cur)
      *cur=(int)exp(((double)*cur-origMin)*scalef+toadd);
  }

void WrappingHeightField :: Square()
  {
    heighttype origMin, origMax;
    GetMinMax(origMin, origMax);
    heighttype range=origMax-origMin+1;
    for (heighttype *cur=mHeights; cur!=mHeights+mSquareSize; ++cur)
      {
        double t=((double)(*cur-origMin))/range;
        *cur=(int) (t*t*range+origMin);
      }
  }

void WrappingHeightField :: ScaleDown(const WrappingHeightField &original)
  {
    assert(original.Size()==2*Size());
    for (int y=0; y<Size(); ++y)
      for (int x=0; x<Size(); ++x)
        {
          int tx=x<<1;
          int ty=y<<1;
          Set(x,y, (original(tx,  ty  ) +
                    original(tx+1,ty  ) +
                    original(tx  ,ty+1) +
                    original(tx+1,ty+1) + 2)>>2);
        }
  }

void WrappingHeightField :: AddScaled(const WrappingHeightField &toadd, float weight)
  {
    assert(toadd.Size()==Size());

    iterator toaddpos=toadd.Begin();
    for (heighttype *cur=mHeights; cur!=mHeights+mSquareSize; ++cur, ++toaddpos)
      *cur+=(int)(weight*(*toaddpos)+.5);
  }

void WrappingHeightField :: SqSqFractal(heighttype noisescale, double noisefactor)
  {
    if (mShiftAmt<2)
      {
        for (heighttype *cur=mHeights; cur!=mHeights+mSquareSize; ++cur)
          *cur=Random(1<<20);
      }
    else
      {
        WrappingHeightField lessDetailed(mShiftAmt-1);
        lessDetailed.SqSqFractal(noisescale, noisefactor);
        noisescale = (int) (noisescale/pow(noisefactor, mShiftAmt-1));
        for (int y=0; y<lessDetailed.Size(); ++y)
          for (int x=0; x<lessDetailed.Size(); ++x)
            {
              heighttype h00=lessDetailed(x,y),
                         h10=lessDetailed(x+1,y),
                         h01=lessDetailed(x,y+1),
                         h11=lessDetailed(x+1,y+1);
              int tx=x<<1, ty=y<<1;
              Set(tx,   ty,   (9*h00+3*(h10+h01)+h11+8+Random(noisescale))>>4);
              Set(tx+1, ty,   (9*h10+3*(h00+h11)+h01+8+Random(noisescale))>>4);
              Set(tx,   ty+1, (9*h01+3*(h00+h11)+h10+8+Random(noisescale))>>4);
              Set(tx+1, ty+1, (9*h11+3*(h10+h01)+h00+8+Random(noisescale))>>4);
            }
      }
  }

void WrappingHeightField :: MultiFractal(const WrappingHeightField &roughness)
  {
    assert(roughness.Size()==Size());
    assert(&roughness != this);
    if (mShiftAmt<2)
      {
        //iterator source=roughness.Begin();
        //for (heighttype *cur=mHeights; cur!=mHeights+mSquareSize; ++cur, ++source)
        //  *cur=Random(*source);
        for (heighttype *cur=mHeights; cur!=mHeights+mSquareSize; ++cur)
          *cur=Random(1<<20);
      }
    else
      {
        WrappingHeightField lessDetailedHeight(mShiftAmt-1);
        {
          WrappingHeightField lessDetailedRoughness(mShiftAmt-1);
          lessDetailedRoughness.ScaleDown(roughness);
          lessDetailedHeight.MultiFractal(lessDetailedRoughness);
        }
        // int divby = pow(noisefactor, mShiftAmt);

        for (int y=0; y<lessDetailedHeight.Size(); ++y)
          for (int x=0; x<lessDetailedHeight.Size(); ++x)
            {
              heighttype h00=lessDetailedHeight(x,y),
                         h10=lessDetailedHeight(x+1,y),
                         h01=lessDetailedHeight(x,y+1),
                         h11=lessDetailedHeight(x+1,y+1);
              int tx=x<<1, ty=y<<1;
              Set(tx,   ty,   (9*h00+3*(h10+h01)+h11+8+Random(roughness(tx  ,ty  ))/mSize)>>4);
              Set(tx+1, ty,   (9*h10+3*(h00+h11)+h01+8+Random(roughness(tx+1,ty  ))/mSize)>>4);
              Set(tx,   ty+1, (9*h01+3*(h00+h11)+h10+8+Random(roughness(tx  ,ty+1))/mSize)>>4);
              Set(tx+1, ty+1, (9*h11+3*(h10+h01)+h00+8+Random(roughness(tx+1,ty+1))/mSize)>>4);
            }
      }
  }

void WrappingHeightField :: JFractal(heighttype noisevaryscale, double noisevaryfactor, 
                                     heighttype minroughness, heighttype maxroughness)
  {
    WrappingHeightField::heighttype minH, maxH;

    // Generate roughness
    WrappingHeightField roughness(mShiftAmt);
    roughness.SqSqFractal(noisevaryscale, noisevaryfactor);

    // smooth roughness
    WrappingHeightField smoothed(mShiftAmt);
    smoothed.SetToSmoothed(roughness);
    smoothed.GetMinMax(minH, maxH);
    int smoothnessrange=(maxH-minH+1);

    // Convert roughness to the right curve
    roughness.Exp(minroughness, maxroughness);
    MultiFractal(roughness);
    GetMinMax(minH, maxH);
    int hfrange=(maxH-minH+1);
    // Add smoothed version to heightfield, with equal weight
    AddScaled(smoothed, ((float) hfrange)/((float) smoothnessrange));
  }

void WrappingHeightField :: ColorJFractal(heighttype noisevaryscale, double noisevaryfactor, 
                                     heighttype minroughness, heighttype maxroughness,
                                     WrappingHeightField & color)
  {
    WrappingHeightField::heighttype minH, maxH;

    // Generate roughness
    WrappingHeightField roughness(mShiftAmt);
    roughness.SqSqFractal(noisevaryscale, noisevaryfactor);
    color=roughness;
    color.Normalize(0, 65535);

    // smooth roughness
    WrappingHeightField smoothed(mShiftAmt);
    smoothed.SetToSmoothed(roughness);
    smoothed.GetMinMax(minH, maxH);
    int smoothnessrange=(maxH-minH+1);

    // Convert roughness to the right curve
    roughness.Exp(minroughness, maxroughness);
    MultiFractal(roughness);
    GetMinMax(minH, maxH);
    int hfrange=(maxH-minH+1);
    // Add smoothed version to heightfield, with equal weight
    AddScaled(smoothed, ((float) hfrange)/((float) smoothnessrange));
  }

#if 0
void WrappingHeightField :: SetToSmoothed(const WrappingHeightField &hf, double strength)
  {
    assert(hf.Size()==Size());
    assert(&hf != this);
    if (mShiftAmt<2)
      {
        iterator source=hf.Begin();
        heighttype avg= (hf(0,0)+hf(1,0)+hf(0,1)+hf(1,1)+2)>>2;
        for (heighttype *cur=mHeights; cur!=mHeights+mSquareSize; ++cur, ++source)
          *cur=(int)(*source*(1.0-strength)+strength*avg);
      }
    else
      {
        WrappingHeightField smallersmoothed(mShiftAmt-1);
        WrappingHeightField scaledDown(mShiftAmt-1);
        scaledDown.ScaleDown(hf);
        smallersmoothed.SetToSmoothed(scaledDown, strength/1.414);
        for (int y=0; y<scaledDown.Size(); ++y)
          for (int x=0; x<scaledDown.Size(); ++x)
            {
              int tx=x<<1, ty=y<<1;
              heighttype h00=hf(tx,ty),
                         h10=hf(tx+1,ty),
                         h01=hf(tx,ty+1),
                         h11=hf(tx+1,ty+1);
              heighttype sum=h00+h10+h01+h11+2;
              heighttype avg00=(sum-h11+hf(tx-1,y)+hf(tx,ty-1))/5,
                         avg01=(sum-h10+hf(tx-1,y)+hf(tx,ty+1))/5,
                         avg10=(sum-h01+hf(tx+1,y)+hf(tx,ty-1))/5,
                         avg11=(sum-h00+hf(tx+1,y)+hf(tx,ty+1))/5;

              heighttype diff=scaledDown(x,y)-smallersmoothed(x,y),
                         diffx0=(scaledDown(x-1,y)-smallersmoothed(x-1,y)+2)>>2,
                         diffx1=(scaledDown(x+1,y)-smallersmoothed(x+1,y)+2)>>2,
                         diffy0=(scaledDown(x,y-1)-smallersmoothed(x,y-1)+2)>>2,
                         diffy1=(scaledDown(x,y+1)-smallersmoothed(x,y+1)+2)>>2;

              Set(tx,   ty,   (int)(h00+diff+diffx0+diffy0+(avg00-h00)*strength));
              Set(tx+1, ty,   (int)(h10+diff+diffx1+diffy0+(avg10-h10)*strength));
              Set(tx,   ty+1, (int)(h01+diff+diffx0+diffy1+(avg01-h01)*strength));
              Set(tx+1, ty+1, (int)(h11+diff+diffx1+diffy1+(avg11-h11)*strength));
            }
      }
  }
#else
void WrappingHeightField :: SetToSmoothed(const WrappingHeightField &hf, int width)
  {
    assert(hf.Size()==Size());
    assert(&hf != this);

    if (width < 0)
      width=(int) sqrt(hf.Size());
    double support=(1+width)*(1+width);
    double sum=0.0;
    int x,y;
    for (y= -width; y<=width; ++y)
      for (x= -width; x<=width; ++x)
        sum += hf(x,y);
    for (x=0; x<hf.Size(); ++x)
      {
        for (y=0; y<hf.Size(); ++y)
          {
            Set(x,y, ((int) (sum/support)));
            for (int i=-width; i<=width; ++i)
              {
                sum += hf(x+i, y+width+1);
                sum -= hf(x+i, y-width);
              }
          }
        for (int i=-width; i<=width; ++i)
          {
            sum += hf(x+width+1, y+i);
            sum -= hf(x-width, y+i);
          }
      }

  }
#endif