// copyright 2002 Josh Levenberg, please see copying.txt
/*
  This implements my new algorithm, "CABTT" 
  (Cached Agregate Binary Triangle Trees)

  Output has fixed quality, rather than fixed triangle count

  Has two types of BTTs: subdivisions of an aggregate triangle: "High detail"
  and the tree of aggregate triangles: "Low detail"

   High detail
   -----------
   Fixed boundary
   Split only
   Can use either adaptive or uniform subdivision
   Rendered once into a cache (see display.h) and then discarded
   [max] depth = 2 * gAggrErrors->GetAggregation()
   Managed by AggrErrorsT (see model.cpp)

   Low detail
   ----------
   Arbitrary boundary
   Split and merge
   Maintained across frames
   Has a split and merge queue
   Stores display list # in leaves
     If 0, needs to be rendered and cached this frame
   No restrictions on BTT structure
   max depth = gModel->GetMaxSubdivision() - 2 * gAggrErrors->GetAggregation()
   Implemented in this file

  Note that the new algorithm combined with a fixed low detail tree
  is used to implement the full mesh drawing case.
*/

#include "display.h"
#include "dlod.h"
#include "cachehlp.h"
#include "model.h"
#include "frustum.h"
// standard C
#include <assert.h>
#include <stdio.h>
//#include <math.h>
// C++ STL stuff
#include <set>
#include <algorithm>

using std::max;
using std::min;

namespace CacheHelper
  {
    bool gIsCaching=false;
    Coord *gCachedVerts;
    Index *gCachedIndex;
    Coord *gCurVert;
    Index *gCurIndex;
  }

using namespace CacheHelper;

static int gOutputTriangles=0;

// aggregation specific globals ------------------------
    
static int gAggregation = 4;
static bool gAdaptive = true;

// -----------------------------------

class BTTTri;
class LookAtQueue;

struct BTTLeaf
  {
    int mVertices[2];        // left and right vertices
                             // can get base vertex from mParent->mVertIndex
    Cache mDisplayList;  // VideoCache::sInvalid if hasn't been compiled yet
    int mNumTriangles;       // only valid if mDisplayList is valid
  };

struct BTTBranch
  {
    int      mNegative;   // negative if a branch, >=0 if a leaf
    BTTTri *mChildren;   // mChildren[0] is left, mChildren[1] is right
  };

class BTTTri
  {
public:
    enum { Left, Right, Base };

protected:
    union
      {
        BTTLeaf   uLeaf;
        BTTBranch uBranch;  // Invariant: IsLeaf() || IsDummyBranch() ||
                             //      (uBranch.mChildren[0].mParent==this &&
                             //       uBranch.mChildren[1].mParent==this)
      };
    BTTTri *mParent;
    BTTTri *mNeighbors[3];  
    // invariant: if this is a leaf, mNeighbors always points to other leaves
    // only mNeighbors[2] is used in Branch nodes
    int      mHypotVert;  // vertex in the midpoint of base edge (hypotenuse)
                          // can use to get radius of isosphere, -1 if undefined
    vec_t    mDistanceUntilLookAt;  // used by LookAtQueue
    // invariant: mDistanceUntilLookAt==InvalidDistance unless
    // this BTTTri is in the split or merge queue
    static const vec_t InvalidDistance;

public:
    // Leaf functions
    bool IsLeaf() const { return uBranch.mNegative>=0; }
    void SetLeaf(int vertices[2])  
                  { assert(vertices[0]>=0); 
                    assert(vertices[1]>=0);
                    uLeaf.mVertices[0]=vertices[0];
                    uLeaf.mVertices[1]=vertices[1];
                    uLeaf.mDisplayList=AbstractDisplay::InvalidCache;
                  }
    int & Vertex(int i) 
                  { assert(IsLeaf()); 
                    assert(i>=0);
                    assert(i<2);
                    return uLeaf.mVertices[i]; }
    Cache & DisplayList()
                  { assert(IsLeaf()); 
                    return uLeaf.mDisplayList;
                  }
    ~BTTTri()     
                  { if (IsLeaf() && AbstractDisplay::IsValidCache(DisplayList()))
                      gDisplay->DiscardCache(DisplayList()); }
    int & TriangleCount()
                  { assert(IsLeaf()); 
                    return uLeaf.mNumTriangles;
                  }

    // Branch Functions
    void SetDummyBranch() // used only by BTTTree
                  { uBranch.mNegative=-1;
                    uBranch.mChildren=0; }
    bool IsDummyBranch()
                  { assert(!IsLeaf());
                    return uBranch.mChildren==0; }
    void SetBranch(BTTTri *children)
                  { assert(children);
                    uBranch.mNegative=-1;
                    uBranch.mChildren=children; }
    BTTTri & Child(int i)
                  { assert(!IsLeaf());
                    assert(i>=0);
                    assert(i<2);
                    assert(uBranch.mChildren);
                    return uBranch.mChildren[i]; }
    BTTTri * GetChildren()
                  { assert(!IsLeaf());
                    assert(uBranch.mChildren);
                    return uBranch.mChildren; }
    inline bool CanBeMerged();

    // Common accessor functions
    BTTTri * Neighbor(int i) // if branch node, only use Base neighbor
                  { assert(i>=0);
                    assert(i<3);
                    assert(mNeighbors[i]==0 || IsLeaf() == mNeighbors[i]->IsLeaf());
                    assert(i==Base || IsLeaf());
                    return mNeighbors[i]; }
    void SetNeighbor(int i, BTTTri *tri) // if branch node, only use Base neighbor
                  { assert(i>=0);
                    assert(i<3);
                    assert(tri==0 || IsLeaf() == tri->IsLeaf());
                    assert(i==Base || IsLeaf());
                    mNeighbors[i]=tri; }
    int & HypotVertIndex() 
                  { return mHypotVert; }
    BTTTri *& Parent()
                  { assert(mParent);
                    return mParent; }
    bool IsInQueue() const
                  { return mDistanceUntilLookAt!=InvalidDistance; }
    void SetNotInQueue()
                  { mDistanceUntilLookAt=InvalidDistance; }
    vec_t &DistanceUntilLookAt()  // used by LookAtQueue
                  { return mDistanceUntilLookAt; }
    vec_t DistanceUntilLookAt() const // used by LookAtQueue
                  { assert(IsInQueue());
                    return mDistanceUntilLookAt; }

    //-- begin/end rendering
    void Render(int base);
    void RenderCrossing(int left, int right, int base, int culling, int which=CheckAll); 
    // Split & Merge
    void Split(bool removefromqueue);
    bool Merge();

    // Recursively free caches
    void FreeCaches();

#ifndef NDEBUG
    void CheckIsGood();
#endif
protected:
    void SplitHelper();
    void MergeHelper1(int whichchild);
    void MergeHelper2();
    void RemoveDiamondFromMerge();
  };

const vec_t BTTTri::InvalidDistance = -65536.;

// should only merge if base neighbor is at same LOD
// if merged, should check to see if more merges can be performed
// (ie, undoing a forced split)

// Triangle memory pool management ---------------------------

static int gNumTriPairs;

static inline void InitTriPool()
  {
    gNumTriPairs=0;
  }

static inline BTTTri *AllocateTwoTris()
  {
    ++gNumTriPairs;
    return new BTTTri[2];
  }

static inline void DeallocateTwoTris(BTTTri *tri)
  {
    --gNumTriPairs;
    delete[] tri;
  }

static inline void CloseTriPool()
  {
    gNumTriPairs=0;
  }

// Priority re-evaluation queue --------------------

class LookAtQueue
  {
protected:
    // Would probably prefer a skip list!
    struct compare
      {
        bool operator()(const BTTTri *l, const BTTTri *r) const
          { return l->DistanceUntilLookAt() < r->DistanceUntilLookAt(); }
      };
    typedef std::multiset<BTTTri *, compare> queue_type;
    typedef queue_type::iterator iter_type;
    queue_type mQueue;
public:
    void  DecrementDistances(vec_t dec)
      { // doesn't change relative order when updated in this way
        for (iter_type cur=mQueue.begin(); cur!=mQueue.end(); ++cur)
          (*cur)->DistanceUntilLookAt() -= dec;
      }
    void Clear()
      {
        mQueue.clear();
      }
    BTTTri *PopIfLessThan(vec_t dist)
      {
        if (! mQueue.empty())
          {
            iter_type beg=mQueue.begin();
            if ((*beg)->DistanceUntilLookAt()<dist)
              {
                BTTTri *ret=*beg;
                mQueue.erase(beg);
                return ret;
              }
          }
        return 0;
      }
    void Enqueue(BTTTri *tri, vec_t dist)
      {
        tri->DistanceUntilLookAt()=dist;
        mQueue.insert(tri);
      }
    void Remove(BTTTri *tri)
      {
        std::pair<iter_type,iter_type> where=mQueue.equal_range(tri);
        assert(where.first!=where.second);
        while (*where.first != tri)
          {
            ++where.first;
            assert(where.first!=where.second);
          }
        mQueue.erase(where.first);
      }
// DEBUG CODE
    bool IsIn(BTTTri *tri)
      {
        std::pair<iter_type,iter_type> where=mQueue.equal_range(tri);
        while (where.first!=where.second)
          {
            if (*where.first == tri)
              return true;
            ++where.first;
          }
        return false;
      }
  };

static LookAtQueue gSplitQueue;  // invariant: only contains pointers to leaf BTTTris
static LookAtQueue gMergeQueue;  // invariant: only contains pointers to branch BTTTris

static vec_t gDistanceSoFar = -1024.;
static vec3_t gLastLocation;


static vec_t ComputeSplitDistanceUntilLookAt(int vertindex)
  {
    vec3_t spherecenter;
    gAggrErrors->GetPosition(vertindex, spherecenter);
    return gDistanceSoFar+
            (sqrt(vdistsq(gCamera[0].mLocation,spherecenter))-
             gAggrErrors->GetRadius(vertindex));
  }

static vec_t ComputeMergeDistanceUntilLookAt(int vertindex)
  {
    vec3_t spherecenter;
    gAggrErrors->GetPosition(vertindex, spherecenter);
    return gDistanceSoFar-
            (sqrt(vdistsq(gCamera[0].mLocation,spherecenter))-
             gAggrErrors->GetRadius(vertindex));
  }

void BTTTri :: RemoveDiamondFromMerge()
  { // Checks if either this or base neighbor is in the merge
    // queue -- and if so, removes it..
    assert(!IsLeaf());
    if (!IsInQueue())
      {
        BTTTri *base=Neighbor(Base);
        if (base && base->IsInQueue())
          { // remove base from merge queue
            gMergeQueue.Remove(base);
            base->SetNotInQueue();
          }
      }
    else
      { // remove this from merge queue
        gMergeQueue.Remove(this);
        SetNotInQueue();
      }
  }

inline bool BTTTri :: CanBeMerged()
  {
    return (Child(Left).IsLeaf() && Child(Right).IsLeaf() &&
        (! Neighbor(Base) || (Neighbor(Base)->Child(Left).IsLeaf() &&
                              Neighbor(Base)->Child(Right).IsLeaf())));
  }


// DEBUG CODE
#ifndef NDEBUG
static int gDebugCount=0;
void BTTTri :: CheckIsGood()
  {
    static int isbad=0;
    if (!(IsLeaf() || IsDummyBranch() ||
          (uBranch.mChildren[0].mParent==this &&
           uBranch.mChildren[1].mParent==this)))
      ++isbad;
    else if (!IsLeaf() && IsInQueue() && ! CanBeMerged())
      ++isbad;
    else if (!IsLeaf() && IsInQueue() && ! gMergeQueue.IsIn(this))
      ++isbad;
    else if (IsLeaf() && IsInQueue() && ! gSplitQueue.IsIn(this))
      ++isbad;
    else if (Neighbor(Base) && Neighbor(Base)->Neighbor(Base)==this &&
              IsInQueue() && Neighbor(Base)->IsInQueue())
      ++isbad;
    else if (IsLeaf() && HypotVertIndex()!=Model::InvalidInd && !IsInQueue() && 
              (Neighbor(Base)==0 || 
               Neighbor(Base)->Neighbor(Base)!=this || 
               !Neighbor(Base)->IsInQueue()))
      ++isbad;
    else if (IsLeaf() && Neighbor(Left) && 
            Neighbor(Left)->Neighbor(Right)!=this && 
            Neighbor(Left)->Neighbor(Base)!=this)
      ++isbad;
    else if (IsLeaf() && Neighbor(Right) && 
            Neighbor(Right)->Neighbor(Left)!=this && 
            Neighbor(Right)->Neighbor(Base)!=this)
      ++isbad;
    else if (IsLeaf() && Neighbor(Base) && 
            Neighbor(Base)->Neighbor(Left)!=this && 
            Neighbor(Base)->Neighbor(Right)!=this && 
            Neighbor(Base)->Neighbor(Base)!=this)
      ++isbad;
    else if (! IsLeaf() && ! IsDummyBranch() && CanBeMerged() && ! IsInQueue() &&
              (Neighbor(Base)==0 || 
               !Neighbor(Base)->IsInQueue()))
      ++isbad;
    else if (! IsLeaf() && ! IsDummyBranch())
      {
        Child(0).CheckIsGood();
        Child(1).CheckIsGood();
      }
  }
#endif

// Triangle Splitting ------------------------------

static int gMergeTests=0;
static int gMerges=0;
static int gSplitTests=0;
static int gSplits=0;
static int gForceSplits=0;

void BTTTri::SplitHelper()
  {
    assert(IsLeaf());
    int verts[3]={ Vertex(Right), Parent()->HypotVertIndex(), Vertex(Left) };

    BTTTri *left=Neighbor(Left);
    BTTTri *right=Neighbor(Right);
    BTTTri *child=AllocateTwoTris();
    child[Left].SetLeaf(verts+1);
    child[Right].SetLeaf(verts);
    child[Left].HypotVertIndex()=gAggrErrors->BisectSafe(verts[1], verts[2]); 
    child[Right].HypotVertIndex()=gAggrErrors->BisectSafe(verts[0], verts[1]);

    child[Left].SetNeighbor(Base, left);
    child[Left].SetNotInQueue();
    if (left)
      if (left->Neighbor(Base)==this)
        left->SetNeighbor(Base, & child[Left]);
      else
        {
          assert(left->Neighbor(Right)==this);
          left->SetNeighbor(Right, & child[Left]);
          // add child[Left] to split queue
          // (but first check to see if we are at the maximum refinement)
          if (child[Left].HypotVertIndex()>=0)
            gSplitQueue.Enqueue(& child[Left], 
                      ComputeSplitDistanceUntilLookAt(child[Left].HypotVertIndex()));
        }
    else 
      // add child[Left] to split queue
      // (but first check to see if we are at the maximum refinement)
      if (child[Left].HypotVertIndex()>=0)
        gSplitQueue.Enqueue(& child[Left], 
                ComputeSplitDistanceUntilLookAt(child[Left].HypotVertIndex()));


    child[Left].SetNeighbor(Left, & child[Right]);

    child[Right].SetNeighbor(Base, right);
    child[Right].SetNotInQueue();
    if (right)
      if (right->Neighbor(Base)==this)
        right->SetNeighbor(Base, & child[Right]);
      else
        {
          assert(right->Neighbor(Left)==this);
          right->SetNeighbor(Left, & child[Right]);
          // add child[Right] to split queue
          // (but first check to see if we are at the maximum refinement)
          if (child[Right].HypotVertIndex()>=0)
            gSplitQueue.Enqueue(& child[Right], 
                      ComputeSplitDistanceUntilLookAt(child[Right].HypotVertIndex()));
        }
    else 
      // add child[Right] to split queue
      // (but first check to see if we are at the maximum refinement)
      if (child[Right].HypotVertIndex()>=0) 
        gSplitQueue.Enqueue(& child[Right], 
                  ComputeSplitDistanceUntilLookAt(child[Right].HypotVertIndex()));
    child[Right].SetNeighbor(Right, & child[Left]);

    child[Left].Parent()=this;
    child[Right].Parent()=this;

    if (AbstractDisplay::IsValidCache(DisplayList()))
      {
        gDisplay->DiscardCache(DisplayList());
        DisplayList()=AbstractDisplay::InvalidCache;
      }
    SetBranch(child);
  }

void BTTTri::Split(bool removefromqueue)
  {
    // remove parent diamond from merge queue
    ++gSplits;
    Parent()->RemoveDiamondFromMerge();

    BTTTri *base=Neighbor(Base);
    if (base && base->Neighbor(Base)!=this)
      { // need to force split base neighbor
        ++gForceSplits;
        Neighbor(Base)->Split(true);
        assert(Neighbor(Base)->Neighbor(Base)==this);
        base=Neighbor(Base);
      }

    SplitHelper();
    if (base)
      {
        base->SplitHelper();
        Child(Left).SetNeighbor(Right, & base->Child(Right));
        base->Child(Right).SetNeighbor(Left, & Child(Left));
        Child(Right).SetNeighbor(Left, & base->Child(Left));
        base->Child(Left).SetNeighbor(Right, & Child(Right));
        base->Parent()->RemoveDiamondFromMerge();
        if (removefromqueue)
          {
            if (IsInQueue())
              gSplitQueue.Remove(this);
            else
              {
                assert(base->IsInQueue());
                gSplitQueue.Remove(base);
                base->SetNotInQueue();
              }
          }
        else
          {
            assert(! base->IsInQueue());
          }
      }
    else
      {
        Child(Left).SetNeighbor(Right, 0);
        Child(Right).SetNeighbor(Left, 0);
        if (removefromqueue)
          gSplitQueue.Remove(this);
      }
    gMergeQueue.Enqueue(this, ComputeMergeDistanceUntilLookAt(HypotVertIndex()));
  }


// Triangle Merging --------------------------------

void BTTTri :: MergeHelper1(int whichchild)
  {
    assert(IsLeaf());

    if (AbstractDisplay::IsValidCache(DisplayList()))
      {
        gDisplay->DiscardCache(DisplayList());
        DisplayList()=AbstractDisplay::InvalidCache;
      }

    // fix up Neighbor(Base) back neighbor to be Parent()
    BTTTri *base=Neighbor(Base);
    if (base)
      {
        if (base->Neighbor(Base)==this)
          {
            base->SetNeighbor(Base, Parent());
            // and get rid of split sphere since was a diamond
            if (IsInQueue())
              {
                gSplitQueue.Remove(this); // remove from split queue
                // don't bother marking NotInQueue since
                // about to delete anyway
                gSplitQueue.Enqueue(base, ComputeSplitDistanceUntilLookAt(base->HypotVertIndex()));
              }
            else
              {
                assert(base->IsInQueue() || base->HypotVertIndex()==-1);
              }
            /* else if (base->IsInQueue())
              { // remove from split queue
                gSplitQueue.Remove(base);
                base->SetNotInQueue();
              }
            */
          }
        else
          { // we know whether we should be the left or right
            // neighbor based on whether we are the left or
            // right child
            assert(base->Neighbor(whichchild)==this);
            base->SetNeighbor(whichchild, Parent());
            assert(IsInQueue() || HypotVertIndex()==-1);
            if (IsInQueue())
              gSplitQueue.Remove(this);
          }
      }
    else if (IsInQueue())
      gSplitQueue.Remove(this); // remove from split queue
  }

void BTTTri :: MergeHelper2()
  {
    int verts[2] = { Child(Left).Vertex(Right),
                     Child(Right).Vertex(Left) };
    BTTTri *child=GetChildren();
    SetLeaf(verts);
    SetNeighbor(Left, child[Left].Neighbor(Base));
    SetNeighbor(Right, child[Right].Neighbor(Base));
    child[Left].MergeHelper1(Right);
    child[Right].MergeHelper1(Left);
    DeallocateTwoTris(child);
  }

bool BTTTri :: Merge()
  {
    assert(! IsLeaf());
    assert(CanBeMerged());
    ++gMerges;
#if 0
    if (! CanBeMerged())
      { // happens in practice.  WHY?
        if (! removefromqueue)
          SetNotInQueue();
        return false;
      }
#endif

    BTTTri *base=Neighbor(Base);

    MergeHelper2();
    
    if (base)
      {
        base->MergeHelper2();
#if 0
        // remove from merge queue?
        if (removefromqueue)
          {
            assert(IsInQueue());
            if (!IsInQueue())
              {
                assert(base->IsInQueue());
                gMergeQueue.Remove(base);
                base->SetNotInQueue();
              }
            else
              { // remove this from merge queue
                gMergeQueue.Remove(this);
                // Don't need to SetNotInQueue();
                // since we're going to put it in the split
                // queue immediately
              }
          }
        else
#endif
          {
            assert(! base->IsInQueue());
          }
      }
    else
      { 
        // remove from merge queue?
        assert(IsInQueue());
#if 0
        if (removefromqueue)
          {
            gMergeQueue.Remove(this);
            // Don't need to SetNotInQueue();
            // since we're going to put it in the split
            // queue immediately
          }
#endif
      }
    // add to split queue
    gSplitQueue.Enqueue(this, ComputeSplitDistanceUntilLookAt(HypotVertIndex()));
    // add parent to merge queue?
    if (Parent()->CanBeMerged())
      {
        gMergeQueue.Enqueue(Parent(), ComputeMergeDistanceUntilLookAt(Parent()->HypotVertIndex()));
      }
    if (base && base->Parent()->CanBeMerged())
      {
        gMergeQueue.Enqueue(base->Parent(), ComputeMergeDistanceUntilLookAt(base->Parent()->HypotVertIndex()));
      }
    return true;
  }

// -------------------------------------------------
// Cache maintainance

// Recursively free caches
void BTTTri :: FreeCaches()
  {
    if (IsLeaf())
      {
        if (AbstractDisplay::IsValidCache(DisplayList()))
          {
            gDisplay->DiscardCache(DisplayList());
            DisplayList()=AbstractDisplay::InvalidCache;
          }
      }
    else
      {
        Child(Left).FreeCaches();
        Child(Right).FreeCaches();
      }
  }

// -------------------------------------------------
// Tree maintainance

// BTTTree extends BTTTri
// It acts as a sentinal, carefully constructed to be the parent of the
// roots of the trees.  Prevents triangles in the base mesh from ever being
// merged.
class BTTTree : private BTTTri
  {
    // invariants:
    // this is always a branch node
    // Child(0) is valid
    // Child(1) is a dummy object designed so that
    //           CanBeMerged (above) always returns false
    //           ie. it is marked as a branch node though
    //               it doesn't have children
    // this is never in the Split or Merge queues
    // it is SetNotInQueue()
    // mHypotVert==base vertex  <-- somewhat counterintuitive
    //                              but necessary so child gets 
    //                              the correct base vertex
    // mParent == 0 (or this?)
    // mNeighbors points to other BTTTree's
    // unlike other branch nodes, mNeighbors[Left] and mNeighbors[Right] 
    //    are maintained accurately (actually, they never change)

    // things that will be accessed through parent pointer:
    // Parent()->HypotVertIndex()         <-- should return the base vertex of child triangle
    // Parent()->RemoveDiamondFromMerge() <-- should do nothing 
    //                             (and will as long as this and Neighbor(Base)
    //                              are SetNotInQueue())
    // Parent()->CanBeMerged()  <-- needs to return false
    int mVerts[2];

public:
    int mLastRendered;
    BTTTree() : mLastRendered(0)
      {
        Parent()=0;
        SetNotInQueue();
        SetBranch(AllocateTwoTris());

        Child(0).SetNotInQueue();
        Child(0).Parent()= this;

        Child(1).SetDummyBranch();
        Child(1).SetNotInQueue();
        Child(1).SetNeighbor(Base, 0);
        Child(1).HypotVertIndex()= -1;
        Child(1).Parent()= this;
      }

    void Render(int culling)
      { 
        Child(0).RenderCrossing(mVerts[0], mVerts[1], HypotVertIndex(), culling); 
      }

    BTTTri *GetRootTriangle()
      { return GetChildren(); }
    // Call InitializeVerts before InitializeNeighbors
    // to avoid tripping an assert
    void InitializeVerts(int verts[3], bool addtoqueue)
      {
        mVerts[0]=verts[0];
        mVerts[1]=verts[1];
        HypotVertIndex()=verts[Base];
        Child(0).SetLeaf(verts);
        Child(0).HypotVertIndex()=gAggrErrors->BisectSafe(verts[0], verts[1]);
        // add Child(0) to Split queue
        if (addtoqueue && Child(0).HypotVertIndex()!=-1)
          gSplitQueue.Enqueue(& Child(0), 
                  ComputeSplitDistanceUntilLookAt(Child(0).HypotVertIndex()));
      }
    void InitializeNeighbors(BTTTree *left, BTTTree *right, BTTTree *base)
      {
        mNeighbors[Left] =left;
        mNeighbors[Right]=right;
        mNeighbors[Base] =base;
        Child(0).SetNeighbor(Left,  left  ? left ->GetRootTriangle() : 0);
        Child(0).SetNeighbor(Right, right ? right->GetRootTriangle() : 0);
        Child(0).SetNeighbor(Base,  base  ? base ->GetRootTriangle() : 0);
      }
// DEBUG CODE
#ifndef NDEBUG
    void CheckTreeGood()
      {
        CheckIsGood();
      }
#endif
  };

static BTTTree *gRoot;

// -------------------------------------------------
// Tree rendering

static int gAggTrisRendered=0;

void BTTTri :: Render(int basevert)
  { 
    if (IsLeaf())
      {
        ++gAggTrisRendered;
        if (AbstractDisplay::IsValidCache(DisplayList()))
          {
            gDisplay->RenderCache(DisplayList());
            // don't know if this is the caching case or not
            //gBandwidth += (1+((TriangleCount()+(3<<(2*gAggregation)))>>1))*12;
          }
        else
          {
            BeginCache();
            // use uniform subdivision at highest detail level
            // (I think adaptive case will always give the same 
            //  result, though maybe not as quickly)
// TODO: TIME BOTH!
            if (HypotVertIndex()==-1)
              {
                TriangleCount()=gAggrErrors->RenderUniform(Vertex(Left), Vertex(Right), basevert);
              }
            else
              {
                TriangleCount()=gAggrErrors->Render(Vertex(Left), Vertex(Right), basevert);
              }
            // vertices * 12 bytes/vertex
            gBandwidth += (1+((TriangleCount()+(3<<(2*gAggregation)))>>1))*12;

            DisplayList()=EndCache(TriangleCount());
            assert(AbstractDisplay::IsValidCache(DisplayList()));
          }
        gOutputTriangles+=TriangleCount();
      }
    else
      {
        Child(Left).Render(HypotVertIndex());
        Child(Right).Render(HypotVertIndex());
      }
  }

void BTTTri :: RenderCrossing(int left, int right, int basevert, int culling, int which)
  {
    if (culling)
      {
        static std::vector<vec3s_t> boundvolume;
        boundvolume.clear();
        gModel->BoundingVolume(gAggrErrors->GetHD(left), 
                              gAggrErrors->GetHD(right), 
                              gAggrErrors->GetHD(basevert), boundvolume);
        int vis = IsVisible(which, boundvolume);
        if (vis==Outside)
          FreeCaches();
        else if (IsLeaf() || vis==Inside)
          Render(basevert);
        else
          {
            Child(Left).RenderCrossing(basevert, left, HypotVertIndex(), culling-1, vis);
            Child(Right).RenderCrossing(right, basevert, HypotVertIndex(), culling-1, vis);
          }
      }
    else
      Render(basevert);
  }

static int offsetamount=1;

static void UpdateGLStateWF1()
  {
    glDisable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT, GL_LINE);
  }

static void UpdateGLStateWF2()
  {
    glPolygonMode(GL_FRONT, GL_LINE);
    glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set wireframe color
  }

static void DrawHelper(void (*statechange)())
  {
    // Render pass
    while (1)
      {
        for (int i=0; i<gModel->GetNumBaseTriangles(); ++i)
          gRoot[i].Render(gDoCulling);

        gBandwidth+=gOutputTriangles*6;
        gNumOutputTris+=gOutputTriangles;
        gOutputTriangles=0;

        if (statechange)
          {
            statechange();
            statechange=0;
          }
        else
          break;
      }
  }

static void DrawBTT(vec3_t location, void (*statechange)())
  {
    // compute distance travelled
    gDistanceSoFar += sqrt(vdistsq(location, gLastLocation));
    gLastLocation[0]=location[0];
    gLastLocation[1]=location[1];
    gLastLocation[2]=location[2];
    
    // need to reset to avoid precision loss?
    if (gDistanceSoFar>1024)
      {
        gDistanceSoFar -= 2048;
        gMergeQueue.DecrementDistances(2048);
        gSplitQueue.DecrementDistances(2048);
      }

    // merge pass
    BTTTri *tri=gMergeQueue.PopIfLessThan(gDistanceSoFar);
    while (tri)
      {
        vec_t dist=ComputeMergeDistanceUntilLookAt(tri->HypotVertIndex());
        ++gMergeTests;
        if (dist<gDistanceSoFar)
          tri->Merge();
        else
          gMergeQueue.Enqueue(tri, dist);
        tri=gMergeQueue.PopIfLessThan(gDistanceSoFar);
      }
    // DEBUG CODE
#ifndef NDEBUG
    {
      for (int j=0; j<gModel->GetNumBaseTriangles(); ++j)
        {
          ++gDebugCount;
          gRoot[j].CheckTreeGood();
        }
    }
#endif 

    // split pass
    tri=gSplitQueue.PopIfLessThan(gDistanceSoFar);
    while (tri)
      {
        vec_t dist=ComputeSplitDistanceUntilLookAt(tri->HypotVertIndex());
        ++gSplitTests;
        if (dist<gDistanceSoFar)
          tri->Split(false);
        else
          gSplitQueue.Enqueue(tri, dist);

        tri=gSplitQueue.PopIfLessThan(gDistanceSoFar);
      }
    // DEBUG CODE
//#ifndef NDEBUG
#if 0
    {
      for (int j=0; j<gModel->GetNumBaseTriangles(); ++j)
        {
          ++gDebugCount;
          gRoot[j].CheckTreeGood();
        }
    }
#endif

    DrawHelper(statechange);
  }

// registered callbacks ------------------------

void DrawNewAlgo(void)
  {
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    static vec3_t up[3]={{0,0,1}, {0,1,0}, {0,1,0} };
    gluLookAt(gCamera[gWhichCamera].mLocation[0], // eye point
              gCamera[gWhichCamera].mLocation[1],
              gCamera[gWhichCamera].mLocation[2],
              gCamera[gWhichCamera].mLookAt[0],   // center point
              gCamera[gWhichCamera].mLookAt[1],
              gCamera[gWhichCamera].mLookAt[2],
              up[gWhichCamera][0],                // up vector
              up[gWhichCamera][1],
              up[gWhichCamera][2]);

    UpdateClipPlanes();
    if (0 /* gWhichCamera */)
      {
        glLineWidth(1);
        glDisable(GL_LINE_SMOOTH);
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        glDisable(GL_TEXTURE_2D);
        glPolygonMode(GL_FRONT, GL_LINE);
        glDisable(GL_POLYGON_OFFSET_LINE);
        glColor4f(.6f, .6f, .6f, 1.0f);  // sets wireframe color
        glLineWidth(4);
        //glEnable(GL_LINE_SMOOTH);
         
        double xtan = tan(gCamera[0].mFov/2)*gWidth/gHeight;
        float vx = gCamera[0].mLookAt[0] - gCamera[0].mLocation[0];
        float vy = gCamera[0].mLookAt[1] - gCamera[0].mLocation[1];
        float tvx = xtan*vx;
        float tvy = xtan*vy;
        glBegin(GL_LINE_STRIP);
        glVertex3f(gCamera[0].mLookAt[0]-tvy,gCamera[0].mLookAt[1]+tvx,gCamera[0].mLocation[2]+1);
        glVertex3f(gCamera[0].mLocation[0],gCamera[0].mLocation[1],gCamera[0].mLocation[2]+1);
        glVertex3f(gCamera[0].mLookAt[0]+tvy,gCamera[0].mLookAt[1]-tvx,gCamera[0].mLocation[2]+1);
        glEnd();
        glPopAttrib();
        glPolygonOffset(-1.0, 0);
      }
    else
      {
        glDisable(GL_LINE_SMOOTH);
        glLineWidth(1);
        glPolygonOffset(-1.0, 0);
        //glEnable(GL_LINE_SMOOTH);
      }

    switch (gWireframe)
      {
        case 0: // textured, not wireframed
          glDisable(GL_POLYGON_OFFSET_LINE);
          glEnable(GL_TEXTURE_2D);
          glPolygonMode(GL_FRONT, GL_FILL); 
          glDisable(GL_POLYGON_OFFSET_LINE);
          break;
        case 1: // textured with wireframes
          glEnable(GL_POLYGON_OFFSET_LINE);
          glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set wireframe color
          break;
        case 2: // white with wireframes
          glDisable(GL_TEXTURE_2D);
          glEnable(GL_POLYGON_OFFSET_LINE);
          glPolygonOffset(0.0, offsetamount);
          break;
        case 3: // only wireframes
          glDisable(GL_TEXTURE_2D);
          glPolygonMode(GL_FRONT, GL_LINE);
          glDisable(GL_POLYGON_OFFSET_LINE);
          //glColor4f(1.0f, 1.0f, 1.0f, 1.0f);  // sets wireframe color
          glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set wireframe color
          break;
      }

    // using display lists slows rendering twice
    // if gWireframe==1 or gWireframe==2
    if (gWireframe==1)
      {
        glEnable(GL_TEXTURE_2D);
        glPolygonMode(GL_FRONT, GL_FILL);

        DrawBTT(gCamera[0].mLocation, UpdateGLStateWF1);
      }
    else if (gWireframe==2)
      {
        glPolygonMode(GL_FRONT, GL_FILL);
        glColor4f(1.0f, 1.0f, 1.0f, 1.0f);  // set fill color

        DrawBTT(gCamera[0].mLocation, UpdateGLStateWF2);
      }
    else
      DrawBTT(gCamera[0].mLocation, 0);
  }

static void InitBaseMesh()
  {
    // initialize base mesh
    int i;

    int numroots=gModel->GetNumBaseTriangles();
    gRoot=new BTTTree[numroots];
    for (i=0; i<numroots; ++i)
      {
        int tempverts[3];
        for (int j=0; j<3; ++j)
          tempverts[j] = gAggrErrors->GetBaseTriangleVert(i, j);
        gRoot[i].InitializeVerts(tempverts, 
                      i<gModel->GetAdjacentBaseTriangle(i, Model::Base));
      }
    for (i=0; i<numroots; ++i)
      {
        BTTTree *nbrs[3];
        for (int j=0; j<3; ++j)
          {
            int tempnbr = gModel->GetAdjacentBaseTriangle(i, j);
            assert(tempnbr<numroots);
            assert(tempnbr>=-1);
            if (tempnbr>=0)
              nbrs[j]=gRoot+tempnbr;
            else
              nbrs[j]=0;
          }
        gRoot[i].InitializeNeighbors(nbrs[0], nbrs[1], nbrs[2]);
      }
    int n = (1<<gAggregation)+1;
          // maxnumverts, maxnumind
    gDisplay->Initialize((n*(n+1))>>1, (n-1)*(n-1)*3);
  }

void InitDisplayNewAlgo()
  {
    fprintf(gOutFile, "Using CABTT drawing method\n");
    printf("Using CABTT drawing method\n"
           "< and > change the aggregation level\n"
           "/ toggles adaptive aggregate triangles\n");

    // initialize position tracking
    gDistanceSoFar = -1024.;
    gLastLocation[0]=gCamera[0].mLocation[0];
    gLastLocation[1]=gCamera[0].mLocation[1];
    gLastLocation[2]=gCamera[0].mLocation[2];

    // initialize triangle pool
    InitTriPool();

    // compute error values
    InitializeAggregation(gAggregation, gAdaptive);

    InitBaseMesh();
  }

static void RecursivelyDeleteTree(BTTTri *todel)
  {
    if (! todel[0].IsLeaf())
      {
        assert(todel[0].GetChildren());
        RecursivelyDeleteTree(todel[0].GetChildren());
      }
    if (! todel[1].IsLeaf() && ! todel[1].IsDummyBranch())  // todel[1] may be a dummy branch
      {
        RecursivelyDeleteTree(todel[1].GetChildren());
      }
   
    DeallocateTwoTris(todel);
  }

void CloseDisplayNewAlgo()
  {
    gSplitQueue.Clear();
    gMergeQueue.Clear();

    // release triangle pool
    for (int i=0; i<gModel->GetNumBaseTriangles(); ++i)
      RecursivelyDeleteTree(gRoot[i].GetRootTriangle());
    CloseTriPool();

    gDisplay->Done();

    // free base mesh
    delete[] gRoot;
  }


static void ResetAggregationLevel(int newagg, bool adaptive)
  {
    gSplitQueue.Clear();
    gMergeQueue.Clear();

    // release triangle pool
    for (int i=0; i<gModel->GetNumBaseTriangles(); ++i)
      RecursivelyDeleteTree(gRoot[i].GetRootTriangle());
    CloseTriPool();

    gDisplay->Done();

    // free base mesh
    delete[] gRoot;

    gAggregation=newagg;
    gAdaptive=adaptive;    

    // initialize position tracking
    gDistanceSoFar = -1024.;
    gLastLocation[0]=gCamera[0].mLocation[0];
    gLastLocation[1]=gCamera[0].mLocation[1];
    gLastLocation[2]=gCamera[0].mLocation[2];

    // initialize triangle pool
    InitTriPool();

    // compute error values
    InitializeAggregation(gAggregation, gAdaptive);

    InitBaseMesh();
  }

void PrintStatsNewAlgo()
  {
    fprintf(gOutFile, "%.2f merge tests, %.2f merges, %.2f split tests, %.2f splits, %.2f force splits per frame\n", 
            ((double)gMergeTests)/gNumFrames, 
            ((double)gMerges)/gNumFrames, 
            ((double)gSplitTests)/gNumFrames, 
            ((double)gSplits)/gNumFrames, 
            ((double)gForceSplits)/gNumFrames);
    fprintf(gOutFile, "%.2f aggregate tris per frame, %.2f sub-triangles per aggregate tri\n",
            ((double)gAggTrisRendered)/gNumFrames, 
            ((double)gNumOutputTris)/gAggTrisRendered);
  }

void ResetStatsNewAlgo()
  {
    gMergeTests=0;
    gMerges=0;
    gSplitTests=0;
    gSplits=0;
    gForceSplits=0;
    gAggTrisRendered=0;
  }

void KeyboardNewAlgo(unsigned char key, int x, int y)
  {
    if (key=='>' && gAggregation<gModel->GetMaxSubdivision()/2)
      {
        ResetAggregationLevel(gAggregation+1, gAdaptive);
      }
    else if (key=='<' && gAggregation>0)
      {
        ResetAggregationLevel(gAggregation-1, gAdaptive);
      }
    else if (key=='/')
      {
        ResetAggregationLevel(gAggregation, ! gAdaptive);
      }
    DefaultKeyboard(key, x, y);
  }

// ------------------ Alternative callbacks for "Full Mesh" -----------------------

void DrawFullMesh(void)
  {
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    static vec3_t up[3]={{0,0,1}, {0,1,0}, {0,1,0} };
    gluLookAt(gCamera[gWhichCamera].mLocation[0], // eye point
              gCamera[gWhichCamera].mLocation[1],
              gCamera[gWhichCamera].mLocation[2],
              gCamera[gWhichCamera].mLookAt[0],   // center point
              gCamera[gWhichCamera].mLookAt[1],
              gCamera[gWhichCamera].mLookAt[2],
              up[gWhichCamera][0],                // up vector
              up[gWhichCamera][1],
              up[gWhichCamera][2]);

    UpdateClipPlanes();
    if (gWhichCamera)
      {
        glLineWidth(1);
        glDisable(GL_LINE_SMOOTH);
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        glDisable(GL_TEXTURE_2D);
        glPolygonMode(GL_FRONT, GL_LINE);
        glDisable(GL_POLYGON_OFFSET_LINE);
        glColor4f(.6f, .6f, .6f, 1.0f);  // sets wireframe color
        glLineWidth(4);
        //glEnable(GL_LINE_SMOOTH);
         
        double xtan = tan(gCamera[0].mFov/2)*gWidth/gHeight;
        float vx = gCamera[0].mLookAt[0] - gCamera[0].mLocation[0];
        float vy = gCamera[0].mLookAt[1] - gCamera[0].mLocation[1];
        float tvx = xtan*vx;
        float tvy = xtan*vy;
        glBegin(GL_LINE_STRIP);
        glVertex3f(gCamera[0].mLookAt[0]-tvy,gCamera[0].mLookAt[1]+tvx,gCamera[0].mLocation[2]+1);
        glVertex3f(gCamera[0].mLocation[0],gCamera[0].mLocation[1],gCamera[0].mLocation[2]+1);
        glVertex3f(gCamera[0].mLookAt[0]+tvy,gCamera[0].mLookAt[1]-tvx,gCamera[0].mLocation[2]+1);
        glEnd();
        glPopAttrib();
        glPolygonOffset(-1.0, 0);
      }
    else
      {
        glDisable(GL_LINE_SMOOTH);
        glLineWidth(1);
        glPolygonOffset(-1.0, 0);
        //glEnable(GL_LINE_SMOOTH);
      }

    switch (gWireframe)
      {
        case 0: // textured, not wireframed
          glDisable(GL_POLYGON_OFFSET_LINE);
          glEnable(GL_TEXTURE_2D);
          glPolygonMode(GL_FRONT, GL_FILL); 
          glDisable(GL_POLYGON_OFFSET_LINE);
          break;
        case 1: // textured with wireframes
          glEnable(GL_POLYGON_OFFSET_LINE);
          glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set wireframe color
          break;
        case 2: // white with wireframes
          glDisable(GL_TEXTURE_2D);
          glEnable(GL_POLYGON_OFFSET_LINE);
          glPolygonOffset(0.0, offsetamount);
          break;
        case 3: // only wireframes
          glDisable(GL_TEXTURE_2D);
          glPolygonMode(GL_FRONT, GL_LINE);
          glDisable(GL_POLYGON_OFFSET_LINE);
          //glColor4f(1.0f, 1.0f, 1.0f, 1.0f);  // sets wireframe color
          glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set wireframe color
          break;
      }

    // using display lists slows rendering twice
    // if gWireframe==1 or gWireframe==2
    if (gWireframe==1)
      {
        glEnable(GL_TEXTURE_2D);
        glPolygonMode(GL_FRONT, GL_FILL);

        DrawHelper(UpdateGLStateWF1);
      }
    else if (gWireframe==2)
      {
        glPolygonMode(GL_FRONT, GL_FILL);
        glColor4f(1.0f, 1.0f, 1.0f, 1.0f);  // set fill color

        DrawHelper(UpdateGLStateWF2);
      }
    else
      DrawHelper(0);
  }

static int gSavedAggregation;
static bool gSavedAdaptive;

static void InsureSubdivided(BTTTri *tri, int depth)
  {
    if (depth)
      {
        assert(! tri->IsLeaf());
        InsureSubdivided(& tri->Child(BTTTri::Left), depth-1);
        InsureSubdivided(& tri->Child(BTTTri::Right), depth-1);
      }
    else if (tri->IsLeaf())
      {
        tri->Split(true);
      }
  }

/* First this creates a base mesh as in the new algorithm case, then it 
   subdivides until the mesh is at full resolution (using InsureSubdivided 
   above).  After that, the coarse ("low detail") mesh is never changed.
   Note that we don't want too much aggregation since that hurts culling
   and may exceed video card limits.
*/

void InitDisplayFullMesh()
  {
    fprintf(gOutFile, "Using Full Mesh drawing method\n");
    printf("Using Full Mesh drawing method\n");

    // initialize triangle pool
    InitTriPool();

    // compute error values
    gSavedAggregation = gAggregation;
    gSavedAdaptive = gAdaptive;
    gAggregation = min(6, gModel->GetMaxSubdivision()/2);
    gAdaptive = false;
    InitializeAggregation(gAggregation, gAdaptive);

    InitBaseMesh();
    for (int i=0; i<2*(gModel->GetMaxSubdivision()/2-gAggregation); ++i)
      for (int r=0; r<gModel->GetNumBaseTriangles(); ++r)
        InsureSubdivided(gRoot[r].GetRootTriangle(), i);
  }

void CloseDisplayFullMesh()
  {
    gAggregation = gSavedAggregation;
    gAdaptive = gSavedAdaptive;

    gSplitQueue.Clear();
    gMergeQueue.Clear();

    // release triangle pool
    for (int i=0; i<gModel->GetNumBaseTriangles(); ++i)
      RecursivelyDeleteTree(gRoot[i].GetRootTriangle());
    CloseTriPool();

    gDisplay->Done();

    // free base mesh
    delete[] gRoot;
  }

void PrintStatsFullMesh()
  {
    PrintStatsNewAlgo();
  }

void ResetStatsFullMesh()
  {
    ResetStatsNewAlgo();
  }
