// copyright 2002 Josh Levenberg, please see copying.txt
#include "uitype.h"

#ifdef DEMO_MOVIE
#include "dlod.h"
#include "model.h"
#include "os.h"

static double sOrbitRadius = 0.24;
static double sAngle=0.0;
static double sTimeStep=0.009;
static double sGravity=1.5;
static double sTollerance=0.8;
static double sAvgFrameRate=60.;
static double sLastTime;

static float sAvatarHdg = 0;
static float sAvatarVel[3] = {0, 0, 0};

static bool gTitle=true;
//const char *gText=0;
static const char * gKeysToExecute=0;

// idea: take sTimeStep if last frame time = 1 / sAvgFrameRate
// so step = sTimeStep * (frame time) / (sAvgFrameRate)
static void AnimateAvatar(double step)
  {
    if (gPaused || gTitle)
      {
        gAvatar[0] = 0.5+sOrbitRadius*cos(sAngle);
        gAvatar[1] = 0.5+sOrbitRadius*sin(sAngle);
        gAvatar[2] = gModel->LookUpHeight(gAvatar[0], gAvatar[1])+gAvatarHeight;
      }
    else
      {
        sAngle += step*2*PI/24;

        gAvatar[0] = 0.5+sOrbitRadius*cos(sAngle);
        gAvatar[1] = 0.5+sOrbitRadius*sin(sAngle);

        float height = gAvatar[2];
        float ground = gModel->LookUpHeight(gAvatar[0], gAvatar[1]);

        sAvatarHdg = sAngle + PI/2;

        double dv = step*sGravity;
        double dh = dv*step;
        float newheight = height+sAvatarVel[2]*step;

        if (newheight < ground + sTollerance*gAvatarHeight)
          { // way too low
            if (sAvatarVel[2]<0)
              sAvatarVel[2]=0;
            sAvatarVel[2] += dv;
            height = max(height+sAvatarVel[2]*step, 
                          ground + sTollerance*gAvatarHeight);
          }
        else
          {
            float idealheight = ground + gAvatarHeight;
            if (newheight > idealheight)
              if (newheight - dh > idealheight)
                sAvatarVel[2]-=dv; // falling
              else // following
                sAvatarVel[2]-=dv*(newheight-idealheight)/dh;
            else
              if (newheight + dh < idealheight)
                sAvatarVel[2] +=dv; // push up
              else // following
                sAvatarVel[2] +=dv*(idealheight-newheight)/dh;
            height += sAvatarVel[2]*step;
          }

        gAvatar[2] = height;
      }
  }

static int sInstruction=0;
static double sAngleTarget=0;

struct InstructType
  {
    double finalangle;
    double framerate;
    const char *keys;
    const char *wavefile;
  };

static InstructType sInstructs[] =
  {
    { PI*2.3, 70., "",    "cpu.wav" },
    { PI*2.6, 23., "w",   "wireframe.wav" },
    { PI*2.9, 25., "ww2", "topview.wav" },
    { PI*3.2, 20., "w1d", "nocaching.wav" },
    { PI*3.28, 3.7, "d",   "roam.wav" }
  };

static const char * sResetKeys="d";

static void GetNextInstruction()
  {
    if (sInstruction>=sizeof(sInstructs)/sizeof(sInstructs[0]))
      {
        gTitle=true;
        sInstruction=0;
        sAngleTarget=0;
        sAngle=0;
        gKeysToExecute=sResetKeys;
        return;
      }
    InstructType *inst=& sInstructs[sInstruction];

    // Play wav file before simulating keys
    if (inst->wavefile)
      PlayWavFile(FindFile(inst->wavefile));

    SimulateKeys(inst->keys);
    sAngleTarget=inst->finalangle;
    sAvgFrameRate=inst->framerate;    
    ++sInstruction;
  }

void glutIdle()
  {
    double time = GetTimeInMS()/1000.0;
    double dt = time - sLastTime;
    sLastTime = time;

    double step= time / sAvgFrameRate;
    if (step<.5)
      step=.5;
    else if (step>2.0)
      step=2.0;

    AnimateAvatar(sTimeStep * step);

    if (!gTitle)
      {
        //  step = sTimeStep * (frame time) / (sAvgFrameRate)
        if (sAngle>=sAngleTarget)
          GetNextInstruction();
      }

    gCamera[0].mLocation[0] = gAvatar[0];
    gCamera[0].mLocation[1] = gAvatar[1];
    gCamera[0].mLocation[2] = gAvatar[2];

    gCamera[0].mLookAt[0] = gAvatar[0] + cos(sAvatarHdg);
    gCamera[0].mLookAt[1] = gAvatar[1] + sin(sAvatarHdg);
    gCamera[0].mLookAt[2] = gAvatar[2];

    gCamera[1].mLocation[0] = gAvatar[0] + 0.5*cos(sAvatarHdg);
    gCamera[1].mLocation[1] = gAvatar[1] + 0.5*sin(sAvatarHdg);

    gCamera[1].mLookAt[0] = gCamera[1].mLocation[0];
    gCamera[1].mLookAt[1] = gCamera[1].mLocation[1];
    gCamera[1].mLookAt[2] = gAvatar[2];

    glutPostRedisplay();
  }

void UIInit()
  {
    sLastTime = GetTimeInMS()/1000.0;
  }

void UIKey(unsigned char key)
  {
    static double saved = 0.0;
    switch (key)
      {
        case ' ':
          gTitle=false;
          break;

        case '$':
          saved=sAngle;
          break;
        case '#':
          sAngle=saved;
          gPaused=true;
          break;
      }
  }

bool UIPreDraw()
  {
    if (gTitle)
      glClearColor(0.0f, 0.0f, 0.0f, 0.5f);
    else if (gWhichCamera)
      glClearColor(1.0f, 1.0f, 1.0f, 0.5f);
    else
      glClearColor(0.4f, 0.6f, 1.0f, 0.5f);
      
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if (gTitle)
      {
        // Draw a title screen
        glEnable (GL_TEXTURE_2D);
        //glDisable(GL_TEXTURE_2D);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(0,0,0, // eye point
                  0,0,1, // center point
                  0,1,0);// up vector

        glBindTexture (GL_TEXTURE_2D, 2);
        glColor3f(1,1,1);
        glBegin (GL_QUADS);
        glTexCoord2f(1.0, 1.0);  glVertex3f(-1.0, -1.0, 2.0);
        glTexCoord2f(1.0, 0.0);  glVertex3f(-1.0,  1.0, 2.0);
        glTexCoord2f(0.0, 0.0);  glVertex3f( 1.0,  1.0, 2.0);
        glTexCoord2f(0.0, 1.0);  glVertex3f( 1.0, -1.0, 2.0);
        glEnd ();

        glBindTexture(GL_TEXTURE_2D, 1); // Restore
        return true;
      }
    else 
      return false;
  }

void UIPostDraw()
  {
    if (gKeysToExecute)
      {
        SimulateKeys(gKeysToExecute);
        gKeysToExecute=0;
      }
  }

#endif