// copyright 2002 Josh Levenberg, please see copying.txt
/* A heavily optimized (but otherwise fairly standard) ROAM implementation.
   Uses AbstractDisplay interface (see display.h) to take advantage of
   directly writing to video memory on NVidia cards.
   Has a pool memory allocator, bucket based priority queue 
   (approximately O(1) for any insert/remove), used to support fan rendering
   but current method is faster on NVidia cards.

   I used to have a split-only ROAM implementation, but it was filled with
   horrible hacks for speed.  It did not survive a refactoring, since I no
   longer needed it.
*/

#include "dlod.h"
#include "roamstd.h"
#include "model.h"
#include "display.h"
#include "frustum.h"
// standard C
#include <assert.h>
#include <stdio.h>
#include <math.h>
// C++ STL stuff
#include <vector>

using std::max;

// -----------------------------------------------------------------------------------------------------
// Technique : Standard ROAM
// (fixed quality, rather than fixed triangle count)

// 510 tris = (approx) 512 verts & 1530 inds
enum { MaxTris=1022, MaxVertsPerBlock=MaxTris+2, 
                    MaxIndicesPerBlock=3*MaxTris };

class RoamTri;
class LookAtQueueROAM;

struct RoamLeaf
  {
    // what about vertices?
    int mVertices[2];        // left and right vertices
                             // can get base vertex from mParent->mVertIndex
  };

struct RoamBranch
  {
    int      mNegative;   // negative if a branch, >=0 if a leaf
    RoamTri *mChildren;   // mChildren[0] is left, mChildren[1] is right
  };

class RoamTri
  { // 32 bytes
protected:
    enum { Left, Right, Base };

    union
      {
        RoamLeaf   uLeaf;
        RoamBranch uBranch;  // Invariant: IsLeaf() || IsDummyBranch() ||
                             //      (uBranch.mChildren[0].mParent==this &&
                             //       uBranch.mChildren[1].mParent==this)
      };
    RoamTri *mParent;
    RoamTri *mNeighbors[3];  
    // invariant: if this is a leaf, mNeighbors always points to other leaves
    // only mNeighbors[2] is used in Branch nodes
    int      mHypotVert;  // vertex in the midpoint of base edge (hypotenuse)
                          // can use to get radius of isosphere, -1 if undefined
    vec_t    mDistanceUntilLookAt;  // used by LookAtQueue
    // invariant: mDistanceUntilLookAt==InvalidDistance unless
    // this RoamTri is in the split or merge queue
    static const vec_t InvalidDistance;

public:
    // Leaf functions
    bool IsLeaf() const { return uBranch.mNegative>=0; }
    void SetLeaf(int vertices[2])  
                  { assert(vertices[0]>=0); 
                    assert(vertices[1]>=0);
                    uLeaf.mVertices[0]=vertices[0];
                    uLeaf.mVertices[1]=vertices[1];
                  }
    int & Vertex(int i) 
                  { assert(IsLeaf()); 
                    assert(i>=0);
                    assert(i<2);
                    return uLeaf.mVertices[i]; }

    // Branch Functions
    void SetDummyBranch() // used only by RoamTree
                  { uBranch.mNegative=-1;
                    uBranch.mChildren=0; }
    bool IsDummyBranch()
                  { assert(!IsLeaf());
                    return uBranch.mChildren==0; }
    void SetBranch(RoamTri *children)
                  { assert(children);
                    uBranch.mNegative=-1;
                    uBranch.mChildren=children; }
    RoamTri & Child(int i)
                  { assert(!IsLeaf());
                    assert(i>=0);
                    assert(i<2);
                    assert(uBranch.mChildren);
                    return uBranch.mChildren[i]; }
    RoamTri * GetChildren()
                  { assert(!IsLeaf());
                    assert(uBranch.mChildren);
                    return uBranch.mChildren; }
    inline bool CanBeMerged();

    // Common accessor functions
    RoamTri * Neighbor(int i) // if branch node, only use Base neighbor
                  { assert(i>=0);
                    assert(i<3);
                    assert(mNeighbors[i]==0 || IsLeaf() == mNeighbors[i]->IsLeaf());
                    assert(i==Base || IsLeaf());
                    return mNeighbors[i]; }
    void SetNeighbor(int i, RoamTri *tri) // if branch node, only use Base neighbor
                  { assert(i>=0);
                    assert(i<3);
                    assert(tri==0 || IsLeaf() == tri->IsLeaf());
                    assert(i==Base || IsLeaf());
                    mNeighbors[i]=tri; }
    int & HypotVertIndex() 
                  { return mHypotVert; }
    RoamTri *& Parent()
                  { assert(mParent);
                    return mParent; }
    bool IsInQueue() const
                  { return mDistanceUntilLookAt!=InvalidDistance; }
    void SetNotInQueue()
                  { mDistanceUntilLookAt=InvalidDistance; }
    vec_t &DistanceUntilLookAt()  // used by LookAtQueue
                  { return mDistanceUntilLookAt; }
    vec_t DistanceUntilLookAt() const // used by LookAtQueue
                  { assert(IsInQueue());
                    return mDistanceUntilLookAt; }

    //-- rendering
    void Render0();
    void Render1();
    void RenderCrossing(int left, int right, int basevert, int culling, int which);

    // Split & Merge
    void Split(bool removefromqueue);
    bool Merge(/*bool removefromqueue -- always false*/);

#ifndef NDEBUG
    void CheckIsGood();
#endif
protected:
    void SplitHelper();
    void MergeHelper1(int whichchild);
    void MergeHelper2();
    void RemoveDiamondFromMerge();
  };

const vec_t RoamTri::InvalidDistance = -65536.;

// should only merge if base neighbor is at same LOD
// if merged, should check to see if more merges can be performed
// (ie, undoing a forced split)

// Roam Triangle memory pool management ---------------------------

static int gNumTriPairs;

union TriPair
  {
    RoamTri mTris[2];
    TriPair *mNext;
  };

struct AllocatedNode
  {
    AllocatedNode *mNext;
    TriPair *mEnd;
    TriPair *mFree;
    TriPair mData[1];
  };

static AllocatedNode *gTriPool=0;
static TriPair *gFreeList=0;

static void InitTriPool()
  { // initial allocation: 16384
    enum { InitialAllocation=16384 };
    gNumTriPairs=0;
    assert(gTriPool==0);
    assert(gFreeList==0);
    gTriPool=static_cast<AllocatedNode *>(malloc(sizeof(AllocatedNode)+(InitialAllocation-1)*sizeof(TriPair)));
    gTriPool->mNext=0;
    gTriPool->mEnd = & gTriPool->mData[InitialAllocation];
    gTriPool->mFree= & gTriPool->mData[0];
  }

static RoamTri *AllocateTwoTris()
  {
    ++gNumTriPairs;
    if (gFreeList)
      { // pull something off the free list
        RoamTri *ret=gFreeList->mTris;
        gFreeList=gFreeList->mNext;
        return ret;
      }
    else if (gTriPool->mFree==gTriPool->mEnd)
      { // allocate more space
        int newsize=gTriPool->mEnd-gTriPool->mData;
        newsize += newsize >> 1;
        AllocatedNode *morespace=static_cast<AllocatedNode *>(malloc(sizeof(AllocatedNode)+(newsize-1)*sizeof(TriPair)));
        morespace->mNext=gTriPool;
        morespace->mEnd = & morespace->mData[newsize];
        morespace->mFree= & morespace->mData[0];
        gTriPool=morespace;
      }
    // pull from the end of the currently allocated node
    return (gTriPool->mFree++)->mTris;
  }

static inline void DeallocateTwoTris(RoamTri *tri)
  {
    --gNumTriPairs;
    TriPair *pair=reinterpret_cast<TriPair *>(tri);
    pair->mNext=gFreeList;
    gFreeList=pair;
  }

static void CloseTriPool()
  {
    while (gTriPool)
      {
        AllocatedNode *next=gTriPool->mNext;
        free(gTriPool);
        gTriPool=next;
      }
    gFreeList=0;
    gNumTriPairs=0;
  }

// End ---------------------------------------------

// Priority re-evaluation queue --------------------

// 0000
//static vec_t gDistanceSoFar = -1024.;
static vec_t gDistanceSoFar = -8.;
static vec3_t gLastLocation;

class LookAtQueueROAM
  {
protected:
    typedef std::vector<int>      bucket_type;
    typedef bucket_type::iterator bucket_iter;
    enum { RegularBuckets = 1<<14, 
           // in addition to regular buckets, have one overflow bucket
           AllBuckets = RegularBuckets+1, 
           OverflowBucket = RegularBuckets,
           BucketsMask = RegularBuckets-1 };
    enum { EmptyBucket = -1 };
    bucket_type mBuckets;
    int mFirstBucket;
#define TotalDistance (2.)
#define BucketDistance (TotalDistance/RegularBuckets)
    vec_t mFirstDistance;

    struct Item
      {
        RoamTri *mTri;
        int mNext;
      };
    typedef std::vector<Item>     items_type;
    typedef items_type::iterator  item_iter;
    items_type mItems;
    int mFirstUnusedItem;

    int WhichBucket(vec_t dist) const
      {
        if (dist<=mFirstDistance)
          return mFirstBucket;
        if (dist-mFirstDistance>=TotalDistance)
          return OverflowBucket;
        else
          return (((unsigned)((dist-mFirstDistance)*
                              (RegularBuckets/TotalDistance))) 
                  + mFirstBucket) & BucketsMask;
      }

public:
    void Init();
    void DecrementDistances(vec_t dec);
    void Clear();
    RoamTri *PopIfLessThan(vec_t dist);
    void Enqueue(RoamTri *tri, vec_t dist);
    void Remove(RoamTri *tri);
    bool IsIn(RoamTri *tri) const;

  };

static LookAtQueueROAM sSplitQueue;  // invariant: only contains pointers to leaf RoamTris
static LookAtQueueROAM sMergeQueue;  // invariant: only contains pointers to branch RoamTris


void LookAtQueueROAM :: Init()
  {
    assert(mBuckets.empty());
    mBuckets.reserve(AllBuckets);
    mBuckets.resize(AllBuckets, EmptyBucket);
    mFirstBucket=0;
    mFirstDistance=gDistanceSoFar;
    assert(mItems.empty());
    mFirstUnusedItem=EmptyBucket;
  }

void LookAtQueueROAM :: DecrementDistances(vec_t dec)
  {
    mFirstDistance -= dec;
    int bucketnum;
    for (bucketnum=0; bucketnum < AllBuckets; ++bucketnum)
      {
        int itemnum = mBuckets[bucketnum];
        while (itemnum != EmptyBucket)
          {
            RoamTri *tri=mItems[itemnum].mTri;
            assert(tri->IsInQueue());
            tri->DistanceUntilLookAt() -= dec;
            itemnum = mItems[itemnum].mNext;
          }
      }
    // do to precision issues, items may move to an adjacent bucket
    bucket_type bucketscopy(AllBuckets, EmptyBucket);
    for (bucketnum=0; bucketnum < AllBuckets; ++bucketnum)
      {
        int itemnum = mBuckets[bucketnum];
        while (itemnum != EmptyBucket)
          {
            mBuckets[bucketnum] = mItems[itemnum].mNext;
            int newbucket = WhichBucket(mItems[itemnum].mTri->DistanceUntilLookAt());
            mItems[itemnum].mNext = bucketscopy[newbucket];
            bucketscopy[newbucket] = itemnum;
            itemnum = mBuckets[bucketnum];
          }
      }
    mBuckets.swap(bucketscopy);
  }

void LookAtQueueROAM :: Clear()
  {
    mBuckets.clear();
    mItems.clear();
  }

RoamTri *LookAtQueueROAM :: PopIfLessThan(vec_t dist)
  {
    assert(dist>=mFirstDistance);
    int lastbucket=WhichBucket(dist);
    while (mFirstBucket != lastbucket)
      {
        int item = mBuckets[mFirstBucket];
        if (item == EmptyBucket)
          {
            ++mFirstBucket;
            mFirstDistance += BucketDistance;
            if (mFirstBucket==RegularBuckets)
              {
                mFirstBucket=0;
                // Scan overflow bucket, moving items as necessary
                int *itemnumptr=&mBuckets[OverflowBucket];
                int itemnum=*itemnumptr;
                while (itemnum != EmptyBucket)
                  {
                    Item *itemptr=&mItems[itemnum];
                    int bucket = WhichBucket(itemptr->mTri->DistanceUntilLookAt());
                    if (bucket!=OverflowBucket)
                      { // Move item out of overflow bucket
                        *itemnumptr = itemptr->mNext;
                        itemptr->mNext=mBuckets[bucket];
                        mBuckets[bucket]=itemnum;
                        itemnum=*itemnumptr;
                      }
                    else
                      { // Leave item in overflow bucket
                        itemnumptr = & itemptr->mNext;
                        itemnum=*itemnumptr;
                      }
                  }
              }
            assert(mFirstBucket == (mFirstBucket & BucketsMask));
          }
        else
          {
            Item *itemptr = & mItems[item];
            mBuckets[mFirstBucket] = itemptr->mNext;
            itemptr->mNext = mFirstUnusedItem;
            mFirstUnusedItem = item;
#ifndef NDEBUG
            RoamTri *ret=itemptr->mTri;
            itemptr->mTri=0;
            return ret;
#else
            return itemptr->mTri;
#endif
          }
      }
    int *curitemptr = & mBuckets[mFirstBucket];
    while (*curitemptr != EmptyBucket)
      {
        int itemnum=*curitemptr;
        Item *itemptr = & mItems[itemnum];
        if (itemptr->mTri->DistanceUntilLookAt() < dist)
          {
            *curitemptr = itemptr->mNext;
            itemptr->mNext = mFirstUnusedItem;
            mFirstUnusedItem = itemnum;
#ifndef NDEBUG
            RoamTri *ret=itemptr->mTri;
            itemptr->mTri=0;
            return ret;
#else
            return itemptr->mTri;
#endif
          }
        curitemptr = & itemptr->mNext;
      }
    return 0;
  }
void LookAtQueueROAM :: Enqueue(RoamTri *tri, vec_t dist)
  {
    int bucket = WhichBucket(dist);
    tri->DistanceUntilLookAt()=dist;
    int item=mFirstUnusedItem;
    if (item==EmptyBucket)
      {
        item=mItems.size();
        mItems.push_back();
      }
    else
      {
        assert(mItems[item].mTri==0);
        mFirstUnusedItem = mItems[item].mNext;
      }
    mItems[item].mTri=tri;
    mItems[item].mNext=mBuckets[bucket];
    mBuckets[bucket]=item;
  }

void Quit();
void LookAtQueueROAM :: Remove(RoamTri *tri)
  {
    assert(tri->IsInQueue());
    int bucket = WhichBucket(tri->DistanceUntilLookAt());
    // scan bucket for tri, then scan overflow bucket for tri
    // (should we move items in overflow bucket??)
    while (1)
      {
        int *numptr=&mBuckets[bucket];
        int itemnum=*numptr;
        while (itemnum!=EmptyBucket)
          {
            Item *itemptr=&mItems[itemnum];
            if (itemptr->mTri==tri)
              {
                *numptr = itemptr->mNext;
                itemptr->mNext = mFirstUnusedItem;
#ifndef NDEBUG
                itemptr->mTri=0;
#endif
                mFirstUnusedItem = itemnum;
                return;
              }
            numptr=&itemptr->mNext;
            itemnum=*numptr;
          }
        assert(bucket!=OverflowBucket);
        if (bucket==OverflowBucket)
          Quit();
        bucket=OverflowBucket;
      }
  }
// DEBUG CODE
bool LookAtQueueROAM :: IsIn(RoamTri *tri) const
  {
    int bucket = WhichBucket(tri->DistanceUntilLookAt());
    while (1)
      {
        int item=mBuckets[bucket];
        while (item!=EmptyBucket)
          {
            if (mItems[item].mTri==tri)
              return true;
            item=mItems[item].mNext;
          }
        if (bucket==OverflowBucket)
          return false;
        bucket=OverflowBucket;
      }
  }

#undef TotalDistance
#undef BucketDistance


static vec_t ComputeSplitDistanceUntilLookAt(int vertindex)
  {
    vec3_t spherecenter;
    gModel->GetPosition(vertindex, spherecenter);
    return gDistanceSoFar+
            (sqrt(vdistsq(gCamera[0].mLocation,spherecenter))-
             gErrors->Get(vertindex));
  }

static vec_t ComputeMergeDistanceUntilLookAt(int vertindex)
  {
    vec3_t spherecenter;
    gModel->GetPosition(vertindex, spherecenter);
    return gDistanceSoFar-
            (sqrt(vdistsq(gCamera[0].mLocation,spherecenter))-
             gErrors->Get(vertindex));
  }

void RoamTri :: RemoveDiamondFromMerge()
  { // Checks if either this or base neighbor is in the merge
    // queue -- and if so, removes it..
    assert(!IsLeaf());
    if (!IsInQueue())
      {
        RoamTri *base=Neighbor(Base);
        if (base && base->IsInQueue())
          { // remove base from merge queue
            sMergeQueue.Remove(base);
            base->SetNotInQueue();
          }
      }
    else
      { // remove this from merge queue
        sMergeQueue.Remove(this);
        SetNotInQueue();
      }
  }

inline bool RoamTri :: CanBeMerged()
  {
    return (Child(Left).IsLeaf() && Child(Right).IsLeaf() &&
        (! Neighbor(Base) || (Neighbor(Base)->Child(Left).IsLeaf() &&
                              Neighbor(Base)->Child(Right).IsLeaf())));
  }


// DEBUG CODE
#ifndef NDEBUG
static int gDebugCount=0;
void RoamTri :: CheckIsGood()
  {
    static int isbad=0;
    if (!(IsLeaf() || IsDummyBranch() ||
          (uBranch.mChildren[0].mParent==this &&
           uBranch.mChildren[1].mParent==this)))
      ++isbad;
    else if (!IsLeaf() && IsInQueue() && ! CanBeMerged())
      ++isbad;
    else if (!IsLeaf() && IsInQueue() && ! sMergeQueue.IsIn(this))
      ++isbad;
    else if (IsLeaf() && IsInQueue() && ! sSplitQueue.IsIn(this))
      ++isbad;
    else if (Neighbor(Base) && Neighbor(Base)->Neighbor(Base)==this &&
              IsInQueue() && Neighbor(Base)->IsInQueue())
      ++isbad;
    else if (IsLeaf() && HypotVertIndex()!=-1 && !IsInQueue() && 
              (Neighbor(Base)==0 || 
               Neighbor(Base)->Neighbor(Base)!=this || 
               !Neighbor(Base)->IsInQueue()))
      ++isbad;
    else if (! IsLeaf() && ! IsDummyBranch() && CanBeMerged() && ! IsInQueue() &&
              (Neighbor(Base)==0 || 
               !Neighbor(Base)->IsInQueue()))
      ++isbad;
    else if (! IsLeaf() && ! IsDummyBranch())
      {
        Child(0).CheckIsGood();
        Child(1).CheckIsGood();
      }
  }
#endif

// Triangle Splitting ------------------------------

static int gMergeTests=0;
static int gMerges=0;
static int gSplitTests=0;
static int gSplits=0;
static int gForceSplits=0;

void RoamTri::SplitHelper()
  {
    assert(IsLeaf());
    int verts[3]={ Vertex(Right), Parent()->HypotVertIndex(), Vertex(Left) };

    RoamTri *left=Neighbor(Left);
    RoamTri *right=Neighbor(Right);
    RoamTri *child=AllocateTwoTris();
    child[Left].SetLeaf(verts+1);
    child[Right].SetLeaf(verts);
    child[Left].HypotVertIndex()=gModel->BisectSafe(verts[1], verts[2]); 
    child[Right].HypotVertIndex()=gModel->BisectSafe(verts[0], verts[1]);

    child[Left].SetNeighbor(Base, left);
    child[Left].SetNotInQueue();
    if (left)
      if (left->Neighbor(Base)==this)
        left->SetNeighbor(Base, & child[Left]);
      else
        {
          assert(left->Neighbor(Right)==this);
          left->SetNeighbor(Right, & child[Left]);
          // add child[Left] to split queue
          // (but first check to see if we are at the maximum refinement)
          if (child[Left].HypotVertIndex()>=0)
            sSplitQueue.Enqueue(& child[Left], 
                      ComputeSplitDistanceUntilLookAt(child[Left].HypotVertIndex()));
        }
    else 
      // add child[Left] to split queue
      // (but first check to see if we are at the maximum refinement)
      if (child[Left].HypotVertIndex()>=0)
        sSplitQueue.Enqueue(& child[Left], 
                ComputeSplitDistanceUntilLookAt(child[Left].HypotVertIndex()));


    child[Left].SetNeighbor(Left, & child[Right]);

    child[Right].SetNeighbor(Base, right);
    child[Right].SetNotInQueue();
    if (right)
      if (right->Neighbor(Base)==this)
        right->SetNeighbor(Base, & child[Right]);
      else
        {
          assert(right->Neighbor(Left)==this);
          right->SetNeighbor(Left, & child[Right]);
          // add child[Right] to split queue
          // (but first check to see if we are at the maximum refinement)
          if (child[Right].HypotVertIndex()>=0)
            sSplitQueue.Enqueue(& child[Right], 
                      ComputeSplitDistanceUntilLookAt(child[Right].HypotVertIndex()));
        }
    else 
      // add child[Right] to split queue
      // (but first check to see if we are at the maximum refinement)
      if (child[Right].HypotVertIndex()>=0) 
        sSplitQueue.Enqueue(& child[Right], 
                  ComputeSplitDistanceUntilLookAt(child[Right].HypotVertIndex()));
    child[Right].SetNeighbor(Right, & child[Left]);

    child[Left].Parent()=this;
    child[Right].Parent()=this;
    SetBranch(child);
  }

void RoamTri::Split(bool removefromqueue)
  {
    // remove parent diamond from merge queue
    ++gSplits;
    Parent()->RemoveDiamondFromMerge();

    RoamTri *base=Neighbor(Base);
    if (base && base->Neighbor(Base)!=this)
      { // need to force split base neighbor
        ++gForceSplits;
        Neighbor(Base)->Split(true);
        assert(Neighbor(Base)->Neighbor(Base)==this);
        base=Neighbor(Base);
      }

    SplitHelper();
    if (base)
      {
        base->SplitHelper();
        Child(Left).SetNeighbor(Right, & base->Child(Right));
        base->Child(Right).SetNeighbor(Left, & Child(Left));
        Child(Right).SetNeighbor(Left, & base->Child(Left));
        base->Child(Left).SetNeighbor(Right, & Child(Right));
        base->Parent()->RemoveDiamondFromMerge();
        if (removefromqueue)
          {
            if (IsInQueue())
              sSplitQueue.Remove(this);
            else
              {
                assert(base->IsInQueue());
                sSplitQueue.Remove(base);
                base->SetNotInQueue();
              }
          }
        else
          {
            assert(! base->IsInQueue());
          }
      }
    else
      {
        Child(Left).SetNeighbor(Right, 0);
        Child(Right).SetNeighbor(Left, 0);
        if (removefromqueue)
          sSplitQueue.Remove(this);
      }
    sMergeQueue.Enqueue(this, ComputeMergeDistanceUntilLookAt(HypotVertIndex()));
  }


// Triangle Merging --------------------------------

void RoamTri :: MergeHelper1(int whichchild)
  {
    assert(IsLeaf());
    // fix up Neighbor(Base) back neighbor to be Parent()
    RoamTri *base=Neighbor(Base);
    if (base)
      {
        if (base->Neighbor(Base)==this)
          {
            base->SetNeighbor(Base, Parent());
            // and get rid of split sphere since was a diamond
            if (IsInQueue())
              {
                sSplitQueue.Remove(this); // remove from split queue
                // don't bother marking NotInQueue since
                // about to delete anyway
                sSplitQueue.Enqueue(base, ComputeSplitDistanceUntilLookAt(base->HypotVertIndex()));
              }
            else
              {
                assert(base->IsInQueue() || base->HypotVertIndex()==-1);
              }
            /* else if (base->IsInQueue())
              { // remove from split queue
                sSplitQueue.Remove(base);
                base->SetNotInQueue();
              }
            */
          }
        else
          { // we know whether we should be the left or right
            // neighbor based on whether we are the left or
            // right child
            assert(base->Neighbor(whichchild)==this);
            base->SetNeighbor(whichchild, Parent());
            assert(IsInQueue() || HypotVertIndex()==-1);
            if (IsInQueue())
              sSplitQueue.Remove(this);
          }
      }
    else if (IsInQueue())
      sSplitQueue.Remove(this); // remove from split queue
  }

void RoamTri :: MergeHelper2()
  {
    int verts[2] = { Child(Left).Vertex(Right),
                     Child(Right).Vertex(Left) };
    RoamTri *child=GetChildren();
    SetLeaf(verts);
    SetNeighbor(Left, child[Left].Neighbor(Base));
    SetNeighbor(Right, child[Right].Neighbor(Base));
    child[Left].MergeHelper1(Right);
    child[Right].MergeHelper1(Left);
    DeallocateTwoTris(child);
  }

bool RoamTri :: Merge(/* bool removefromqueue -- always false */)
  {
    assert(! IsLeaf());
    assert(CanBeMerged());
    ++gMerges;
#if 0
    if (! CanBeMerged())
      { // happens in practice.  WHY?
        if (! removefromqueue)
          SetNotInQueue();
        return false;
      }
#endif

    RoamTri *base=Neighbor(Base);

    MergeHelper2();
    
    if (base)
      {
        base->MergeHelper2();
#if 0
        // remove from merge queue?
        if (removefromqueue)
          {
            assert(IsInQueue());
            if (!IsInQueue())
              {
                assert(base->IsInQueue());
                sMergeQueue.Remove(base);
                base->SetNotInQueue();
              }
            else
              { // remove this from merge queue
                sMergeQueue.Remove(this);
                // Don't need to SetNotInQueue();
                // since we're going to put it in the split
                // queue immediately
              }
          }
        else
#endif
          {
            assert(! base->IsInQueue());
          }
      }
    else
      { 
        // remove from merge queue?
        assert(IsInQueue());
#if 0
        if (removefromqueue)
          {
            sMergeQueue.Remove(this);
            // Don't need to SetNotInQueue();
            // since we're going to put it in the split
            // queue immediately
          }
#endif
      }
    // add to split queue
    sSplitQueue.Enqueue(this, ComputeSplitDistanceUntilLookAt(HypotVertIndex()));
    // add parent to merge queue?
    if (Parent()->CanBeMerged())
      {
        sMergeQueue.Enqueue(Parent(), ComputeMergeDistanceUntilLookAt(Parent()->HypotVertIndex()));
      }
    if (base && base->Parent()->CanBeMerged())
      {
        sMergeQueue.Enqueue(base->Parent(), ComputeMergeDistanceUntilLookAt(base->Parent()->HypotVertIndex()));
      }
    return true;
  }


// -------------------------------------------------
// Tree maintainance

// RoamTree extends RoamTri
// It acts as a sentinal, carefully constructed to be the parent of the
// roots of the trees.  Prevents triangles in the base mesh from ever being
// merged.
class RoamTree : private RoamTri
  {
    // invariants:
    // this is always a branch node
    // Child(0) is valid
    // Child(1) is a dummy object designed so that
    //           CanBeMerged (above) always returns false
    //           ie. it is marked as a branch node though
    //               it doesn't have children
    // this is never in the Split or Merge queues
    // it is SetNotInQueue()
    // mHypotVert==base vertex  <-- somewhat counterintuitive
    //                              but necessary so child gets 
    //                              the correct base vertex
    // mParent == 0 (or this?)
    // mNeighbors points to other RoamTree's
    // unlike other branch nodes, mNeighbors[Left] and mNeighbors[Right] 
    //    are maintained accurately (actually, they never change)

    // things that will be accessed through parent pointer:
    // Parent()->HypotVertIndex()         <-- should return the base vertex of child triangle
    // Parent()->RemoveDiamondFromMerge() <-- should do nothing 
    //                             (and will as long as this and Neighbor(Base)
    //                              are SetNotInQueue())
    // Parent()->CanBeMerged()  <-- needs to return false
    int mVerts[2];

public:
    RoamTree()
      {
        Parent()=0;
        SetNotInQueue();
        SetBranch(AllocateTwoTris());

        Child(0).SetNotInQueue();
        Child(0).Parent()= this;

        Child(1).SetDummyBranch();
        Child(1).SetNotInQueue();
        Child(1).SetNeighbor(Base, 0);
        Child(1).HypotVertIndex()= -1;
        Child(1).Parent()= this;
      }

    void Render(int culling);

    RoamTri *GetRootTriangle()
      { return GetChildren(); }
    // Call InitializeVerts before InitializeNeighbors
    // to avoid tripping an assert
    void InitializeVerts(int verts[3], bool addtoqueue)
      {
        mVerts[Left]=verts[Left];
        mVerts[Right]=verts[Right];
        HypotVertIndex()=verts[Base];
        Child(0).SetLeaf(verts);
        Child(0).HypotVertIndex()=gModel->BisectSafe(verts[0], verts[1]);
        // add Child(0) to Split queue
        if (addtoqueue)
          sSplitQueue.Enqueue(& Child(0), 
                  ComputeSplitDistanceUntilLookAt(Child(0).HypotVertIndex()));
      }
    void InitializeNeighbors(RoamTree *left, RoamTree *right, RoamTree *base)
      {
        mNeighbors[Left] =left;
        mNeighbors[Right]=right;
        mNeighbors[Base] =base;
        Child(0).SetNeighbor(Left,  left  ? left ->GetRootTriangle() : 0);
        Child(0).SetNeighbor(Right, right ? right->GetRootTriangle() : 0);
        Child(0).SetNeighbor(Base,  base  ? base ->GetRootTriangle() : 0);
      }
// DEBUG CODE
#ifndef NDEBUG
    void CheckTreeGood()
      {
        CheckIsGood();
      }
#endif
  };

static RoamTree *gRoot;

// -------------------------------------------------
// Tree rendering


static int gOutputTriangles=0;

struct Recent
  {
    int mInd;
    Index mIndex;
  };
static Index sCurIndex;
static vec_t *sOutputVertices;
static vec_t *sOutputVerticesBegin;
static vec_t *sOutputVerticesEnd;
static Index *sOutputIndices;
static Index *sOutputIndicesBegin;
static Index *sOutputIndicesEnd;
static Recent sLast3[3];

static void InitOutput()
  {
    gDisplay->BeginCache(sOutputVerticesBegin, sOutputIndicesBegin);
    sOutputVertices=sOutputVerticesBegin;
    sOutputVerticesEnd=sOutputVerticesBegin+MaxVertsPerBlock;
    sOutputIndices=sOutputIndicesBegin;
    sOutputIndicesEnd=sOutputIndicesBegin+MaxIndicesPerBlock;
    sLast3[0].mInd=Model::InvalidInd;
    sLast3[1].mInd=Model::InvalidInd;
    sLast3[2].mInd=Model::InvalidInd;
    sCurIndex=-1;
  }

static void FlushOutput()
  {
    int numind=sOutputIndices-sOutputIndicesBegin;
    gOutputTriangles += numind/3;
    Cache c=gDisplay->FinishCache((sOutputVertices-sOutputVerticesBegin)/FloatsPerVert, numind);
    gDisplay->DiscardCache(c);
  }

static inline Index WriteInd(int i)
  {
    if (sLast3[0].mInd==i)
      return sLast3[0].mIndex;
    if (sLast3[1].mInd==i)
      return sLast3[1].mIndex;
    if (sLast3[2].mInd==i)
      return sLast3[2].mIndex;
    gModel->GetPosition(i, sOutputVertices);
    sOutputVertices+=FloatsPerVert;
    return ++sCurIndex;
  }

static void OutputTri(int i1, int i2, int i3)
  {
    // if buffer full
    if (sOutputVertices+3*FloatsPerVert >= sOutputVerticesEnd)
      {
        FlushOutput();
        InitOutput();
      }

    // write all but same to sOutputVertices
    // write all three to sOutputIndices
    Index j1, j2, j3;
    j1=WriteInd(i1);
    j2=WriteInd(i2);
    j3=WriteInd(i3);
    sOutputIndices[0]=j1;
    sOutputIndices[1]=j2;
    sOutputIndices[2]=j3;
    sOutputIndices+=3;
    // update recent
    sLast3[0].mInd=i1;
    sLast3[0].mIndex=j1;
    sLast3[1].mInd=i2;
    sLast3[1].mIndex=j2;
    sLast3[2].mInd=i3;
    sLast3[2].mIndex=j3;
  }

void RoamTri :: Render0()
  {
    if (IsLeaf())
      {
        OutputTri(Parent()->HypotVertIndex(), Vertex(Left), Vertex(Right));
      }
    else
      {
        Child(Left).Render1();
        Child(Right).Render1();
      }
  }

void RoamTri :: Render1()
  { 
    if (IsLeaf())
      {
        OutputTri(Parent()->HypotVertIndex(), Vertex(Left), Vertex(Right));
      }
    else
      {
        Child(Right).Render0();
        Child(Left).Render0();
      }
  }

void RoamTri :: RenderCrossing(int left, int right, int basevert, int culling, int which)
  {
    if (culling)
      {
        static std::vector<vec3s_t> boundvolume;
        boundvolume.clear();
        gModel->BoundingVolume(left, right, basevert, boundvolume);
        int vis = IsVisible(which, boundvolume);
        if (vis==Outside)
          return;
        else if (IsLeaf() || vis==Inside)
          /* fall through */;
        else
          {
            Child(Left).RenderCrossing(basevert, left, HypotVertIndex(), culling-1, vis);
            Child(Right).RenderCrossing(right, basevert, HypotVertIndex(), culling-1, vis);
            return;
          }
      }
    // render in this case
    Render0();
  }

void RoamTree::Render(int culling)
  {
    InitOutput();
    Child(0).RenderCrossing(mVerts[0], mVerts[1], HypotVertIndex(), culling, CheckAll);
    FlushOutput();
  }

#if 0 
static int gFanVertIndices[10];
static int gFanCount=0;
static bool gFanReverse;
static int gOutputFansThisFrame=0;
static int gOutputFans=0;

void RoamTri :: OutputFan()
  {
    assert(gFanCount>=2);
    assert(gFanCount<=10);
    if (!gFanReverse)
      {
        int i=1;
        int until = ((gFanCount+1)>>1);
        do 
          {
            std::swap(gFanVertIndices[i], gFanVertIndices[gFanCount-i]);
            ++i;
          } while (i < until);
      }
    glDrawElements(GL_TRIANGLE_FAN, gFanCount, GL_UNSIGNED_INT, gFanVertIndices);
    gOutputTriangles+=gFanCount-2;
    if (gFanCount>2)
      ++gOutputFansThisFrame;
  }

void RoamTri :: RenderFan0L()
  {
    if (IsLeaf())
      {
        int trailing=gModel->LockedInd(Parent()->HypotVertIndex());
        int bestcenter=gModel->LockedInd(Vertex(Left));
        int leading=gModel->LockedInd(Vertex(Right));
        
        if (bestcenter==gFanVertIndices[0] /* fan vertex */)
          {
            if (trailing==gFanVertIndices[gFanCount-1])
              { // add triangle to fan, same level of detail
                gFanVertIndices[gFanCount++]=leading;
                return;
              }
          }
        else if (bestcenter==gFanVertIndices[gFanCount-1])
          {
            if (leading==gFanVertIndices[0])
              { // case I after case B
                gFanVertIndices[gFanCount++]=trailing; // Vertex(Left) or Vertex(Right)
                                                       // whichver is not bestcenter
                return;
              }
            else if (trailing==gFanVertIndices[0])
              { // fan going up a level of detail
                /*  ____
                    |1R/\   Case A
                    | /0L\
                    |/____\
                */
                gFanVertIndices[gFanCount++]=leading;
                return;
              }
          }

        OutputFan(); // does happen

        // *Begin new fan*
        gFanVertIndices[0]=bestcenter;  // ??
        gFanVertIndices[1]=trailing;
        gFanVertIndices[2]=leading;
        gFanCount=3;
        gFanReverse=false;
      }
    else
      {
        Child(Left). RenderFan1L();
        Child(Right).RenderFan1R();
      }
  }

void RoamTri :: RenderFan0R()
  {
    if (IsLeaf())
      {
        int leading=gModel->LockedInd(Parent()->HypotVertIndex()); // v
        int trailing=gModel->LockedInd(Vertex(Left));
        int bestcenter=gModel->LockedInd(Vertex(Right));
        
        if (bestcenter==gFanVertIndices[0] /* fan vertex */)
          {
            if (trailing==gFanVertIndices[gFanCount-1])
              { // add triangle to fan, same level of detail
                gFanVertIndices[gFanCount++]=leading;
                return;
              }
          }
        else if (trailing==gFanVertIndices[0] && 
                 bestcenter==gFanVertIndices[gFanCount-1])
          if (gFanCount==3)
            /* |\
               | \   Case C
               |1L\
               |___\  /|\
                \0R/ / | \
                  * /  |0R>
                   / 1L| /
                  /____|/
            */
            { // switch fan vertex
              std::swap(gFanVertIndices[0], gFanVertIndices[2]);
              gFanVertIndices[gFanCount++]=leading;
              gFanReverse=!gFanReverse;
              return;
            }
          else // Case F after case D
            { 
              gFanVertIndices[gFanCount++]=leading;
              return;
            }
        else if (trailing==gFanVertIndices[gFanCount-1])
          {
            if (leading==gFanVertIndices[0])
              { // Case G after case A
                gFanVertIndices[gFanCount++]=bestcenter;
                return;
              }
            else if (gFanCount==3 &&
                     leading==gFanVertIndices[1])
              { // 1R started, up a level of detail
                /*      /|       ______
                       / |  or   \ 0R /|
                      /  |        \  / |
                     / 0R|         \/  |
                    |\   |          \0L|
                    | \  |           \ |
                    |1R\ |            \|
                    |___\|
                */
                std::swap(gFanVertIndices[0], gFanVertIndices[1]);
                gFanVertIndices[gFanCount++]=bestcenter;
                gFanReverse=!gFanReverse; // always set true?
                return;
              }
          }
        //else if (/* Parent()->HypotVertIndex() */
        //          leading==gFanVertIndices[0] &&
        //          bestcenter==gFanVertIndices[gFanCount-1])
        //  { // fan going up a level of detail
        //    gFanVertIndices[gFanCount++]=trailing;  // Vertex(Left) or Vertex(Right)
        //                                           // whichver is not bestcenter
        //    return;
        //  }
        OutputFan();

        // *Begin new fan*
        gFanVertIndices[0]=bestcenter;  // v
        gFanVertIndices[1]=trailing;
        gFanVertIndices[2]=leading;
        gFanCount=3;
        gFanReverse=false;
      }
    else
      {
        Child(Left). RenderFan1L();
        Child(Right).RenderFan1R();
      }
  }

void RoamTri :: RenderFan1L()
  {
    if (IsLeaf())
      {
        int leading=gModel->LockedInd(Parent()->HypotVertIndex());  // v
        int bestcenter=gModel->LockedInd(Vertex(Left));
        int trailing=gModel->LockedInd(Vertex(Right));
        
        if (bestcenter==gFanVertIndices[0] /* fan vertex */)
          {
            if (trailing==gFanVertIndices[gFanCount-1])
              { // add triangle to fan, same level of detail
                gFanVertIndices[gFanCount++]=leading;
                return;
              }
          }
        else if (trailing==gFanVertIndices[0] && 
                 bestcenter==gFanVertIndices[gFanCount-1])
          if (gFanCount==3)
              /* Case D
                    /\
                   /1L\
                  /____\
                  |    /
                  | 0R/
                  |  /
                  | /
                  |/
              */
            { // switch fan vertex
              std::swap(gFanVertIndices[0], gFanVertIndices[2]);
              gFanVertIndices[gFanCount++]=leading;
              gFanReverse=!gFanReverse; // always set true?
              return;
            }
          else
            /* Case E (after case C)
                      /|\
                     / | \
                    /  |0R>
                   / 1L| /|
                  /____|/_| <- 1L
            */
            {
              gFanVertIndices[gFanCount++]=leading;
              return;
            }
        else if (trailing==gFanVertIndices[gFanCount-1])
          {
            if (leading==gFanVertIndices[0])
              { // Case J after case B or 1R started
                gFanVertIndices[gFanCount++]=bestcenter;
                return;
              }
            else if (gFanCount==3 && 
                     bestcenter==gFanVertIndices[1])
              { /* 1R started the fan and got the direction wrong
                   ____
                   |1R/|
                   | / |
                   |/1L|
                 */
                std::swap(gFanVertIndices[0], gFanVertIndices[1]);
                gFanVertIndices[gFanCount++]=leading;
                gFanReverse=!gFanReverse; // always set true?
                return;
              }
          }
        //else if (/* Parent()->HypotVertIndex() */
        //          trailing==gFanVertIndices[0] &&           // v
        //          bestcenter==gFanVertIndices[gFanCount-1])
        //  { // fan going up a level of detail
        //    gFanVertIndices[gFanCount++]=leading;  // Vertex(Left) or Vertex(Right)
        //                                           // whichver is not bestcenter
        //    return;
        //  }
        //else if (leading==gFanVertIndices[0] &&          // v
        //         leading==gFanVertIndices[gFanCount-1])
        //  { // going down one level of detail
        //    gFanVertIndices[gFanCount++]=leading;
        //    return;
        //  }
        //else if (trailing==gFanVertIndices[0] &&
        //         leading==gFanVertIndices[gFanCount-1])
        //  { /* 1R started the fan
        //        /|\
        //       / | \
        //      /1R|1L\
        //    /____|___\
        //      1R could start the fan the other way, or we could switch right here
        //    */
        //    gFanVertIndices[gFanCount++]=bestcenter; // might not happen anymore
        //    return;
        //  }
        OutputFan();

        // *Begin new fan*
        gFanVertIndices[0]=bestcenter;  // v
        gFanVertIndices[1]=trailing;
        gFanVertIndices[2]=leading;
        gFanCount=3;
        gFanReverse=true;
      }
    else
      {
        Child(Right).RenderFan0R();
        Child(Left). RenderFan0L();
      }
  }

void RoamTri :: RenderFan1R()
  {
    if (IsLeaf())
      {
        int trailing=gModel->LockedInd(Parent()->HypotVertIndex()); // v
        int leading=gModel->LockedInd(Vertex(Left));
        int bestcenter=gModel->LockedInd(Vertex(Right));
        
        if (bestcenter==gFanVertIndices[0] /* fan vertex */)
          {
            if (trailing==gFanVertIndices[gFanCount-1])
              { // add triangle to fan, same level of detail
                gFanVertIndices[gFanCount++]=leading;
                return;
              }
          }
        else if (bestcenter==gFanVertIndices[gFanCount-1])
          {
            if (trailing==gFanVertIndices[0])
              /*   /|\   Case B
                  / | \
                 <0L|  \
                  \ |1R \
                   \|____\
               */
              { // fan going up a level of detail
                gFanVertIndices[gFanCount++]=leading;
                return;
              }
            //else if (leading==gFanVertIndices[0])
            //  { // Case H after case A
            //    gFanVertIndices[gFanCount++]=trailing; // might not happen
            //    return;
            //  }
          }

        OutputFan();

        // *Begin new fan*
        /* Which of these is better?!?!?
        gFanVertIndices[0]=leading;     // v ??
        gFanVertIndices[1]=bestcenter;
        gFanVertIndices[2]=trailing;
        gFanCount=3;
        gFanReverse=true;
        */
        gFanVertIndices[0]=trailing;     // v
        gFanVertIndices[1]=bestcenter;
        gFanVertIndices[2]=leading;
        gFanCount=3;
        gFanReverse=false;
      }
    else
      {
        Child(Right).RenderFan0R();
        Child(Left). RenderFan0L();
      }
  }

void RoamTri :: RenderFanCrossing(int left, int right, int basevert, int culling, int which)
  {
    if (culling)
      {
        static std::vector<vec3s_t> boundvolume;
        boundvolume.clear();
        gModel->BoundingVolume(left, right, basevert, boundvolume);
        int vis = IsVisible(which, boundvolume);
        if (vis==Outside)
          return;
        else if (IsLeaf() || vis==Inside)
          /* fall through */;
        else
          {
            Child(Left).RenderFanCrossing(basevert, left, HypotVertIndex(), culling-1, vis);
            Child(Right).RenderFanCrossing(right, basevert, HypotVertIndex(), culling-1, vis);
            return;
          }
      }
    // render in this case
    // create an degenerate fan
    gFanCount=2;
    gFanVertIndices[0]=0;
    gFanVertIndices[1]=0;
    gFanReverse=false;
    RenderFan0L();
    OutputFan();
  }

void RoamTree :: FanRender(int culling)
  { 
    Child(0).RenderFanCrossing(mVerts[0], mVerts[1], HypotVertIndex(), culling, CheckAll);
  }
#endif

static void UpdateGLStateWF1()
  {
    glDisable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT, GL_LINE);
  }

static void UpdateGLStateWF2()
  {
    glPolygonMode(GL_FRONT, GL_LINE);
    glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set wireframe color
  }

static void DrawRoam(vec3_t location, void (*statechange)())
  {
    // compute distance travelled
    gDistanceSoFar += sqrt(vdistsq(location, gLastLocation));
    gLastLocation[0]=location[0];
    gLastLocation[1]=location[1];
    gLastLocation[2]=location[2];
    
    // need to reset to avoid precision loss?
    // 0000
    //if (gDistanceSoFar>1024)
    if (gDistanceSoFar>8.)
      {
        // 0000
        //gDistanceSoFar -= 2048;
        gDistanceSoFar -= 8;
        sMergeQueue.DecrementDistances(8);
        sSplitQueue.DecrementDistances(8);
      }

    // merge pass
    RoamTri *tri=sMergeQueue.PopIfLessThan(gDistanceSoFar);
    while (tri)
      {
        vec_t dist=ComputeMergeDistanceUntilLookAt(tri->HypotVertIndex());
        ++gMergeTests;
        if (dist<gDistanceSoFar)
          tri->Merge(/* false -- redundant now */);
        else
          sMergeQueue.Enqueue(tri, dist);
        tri=sMergeQueue.PopIfLessThan(gDistanceSoFar);
      }

    // split pass
    tri=sSplitQueue.PopIfLessThan(gDistanceSoFar);
    while (tri)
      {
        vec_t dist=ComputeSplitDistanceUntilLookAt(tri->HypotVertIndex());
        ++gSplitTests;
        if (dist<gDistanceSoFar)
          tri->Split(false);
        else
          sSplitQueue.Enqueue(tri, dist);

        tri=sSplitQueue.PopIfLessThan(gDistanceSoFar);
      }
    // DEBUG CODE
#if 0 // ndef NDEBUG
    {
      for (int j=0; j<gModel->GetNumBaseTriangles(); ++j)
        {
          ++gDebugCount;
          gRoot[j].CheckTreeGood();
        }
    }
#endif

    // Render pass
    while (1)
      {
        int culling=max(0,gDoCulling-3);
        for (int i=0; i<gModel->GetNumBaseTriangles(); ++i)
#if 0
          gRoot[i].FanRender(culling);
        gBandwidth+=(gOutputTriangles+2*gOutputFansThisFrame)*4;
        gOutputFans+=gOutputFansThisFrame;
        gOutputFansThisFrame=0;
#else
          gRoot[i].Render(culling);
        gBandwidth+=gOutputTriangles*18; // approximately
#endif

        gNumOutputTris+=gOutputTriangles;
        gOutputTriangles=0;

        if (statechange)
          {
            statechange();
            statechange=0;
          }
        else
          break;
      }
  }


void DrawRoamStandard(void)
  {
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    static vec3_t up[3]={{0,0,1}, {0,1,0}, {0,1,0} };
    gluLookAt(gCamera[gWhichCamera].mLocation[0], // eye point
              gCamera[gWhichCamera].mLocation[1],
              gCamera[gWhichCamera].mLocation[2],
              gCamera[gWhichCamera].mLookAt[0],   // center point
              gCamera[gWhichCamera].mLookAt[1],
              gCamera[gWhichCamera].mLookAt[2],
              up[gWhichCamera][0],                // up vector
              up[gWhichCamera][1],
              up[gWhichCamera][2]);

    if (gDoCulling)
      UpdateClipPlanes();

    if (0 /* gWhichCamera */)
      {
        glLineWidth(1);
        glDisable(GL_LINE_SMOOTH);
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        glDisable(GL_TEXTURE_2D);
        glPolygonMode(GL_FRONT, GL_LINE);
        glDisable(GL_POLYGON_OFFSET_LINE);
        glColor4f(.6f, .6f, .6f, 1.0f);  // sets wireframe color
        glLineWidth(4);
        //glEnable(GL_LINE_SMOOTH);
         
        double xtan = tan(gCamera[0].mFov/2)*gWidth/gHeight;
        float vx = gCamera[0].mLookAt[0] - gCamera[0].mLocation[0];
        float vy = gCamera[0].mLookAt[1] - gCamera[0].mLocation[1];
        float tvx = xtan*vx;
        float tvy = xtan*vy;
        glBegin(GL_LINE_STRIP);
        glVertex3f(gCamera[0].mLookAt[0]-tvy,gCamera[0].mLookAt[1]+tvx,gCamera[0].mLocation[2]+1);
        glVertex3f(gCamera[0].mLocation[0],gCamera[0].mLocation[1],gCamera[0].mLocation[2]+1);
        glVertex3f(gCamera[0].mLookAt[0]+tvy,gCamera[0].mLookAt[1]-tvx,gCamera[0].mLocation[2]+1);
        glEnd();
        glPopAttrib();
        glPolygonOffset(-1.0, 0);
      }
    else
      {
        glDisable(GL_LINE_SMOOTH);
        glLineWidth(1);
        glPolygonOffset(-1.0, 0);
        //glEnable(GL_LINE_SMOOTH);
      }

    switch (gWireframe)
      {
        case 0: // textured, not wireframed
          glEnable(GL_TEXTURE_2D);
          glPolygonMode(GL_FRONT, GL_FILL); 
          glDisable(GL_POLYGON_OFFSET_LINE);
          break;
        case 1: // textured with wireframes
          glEnable(GL_POLYGON_OFFSET_LINE);
          glPolygonOffset(0.0, -4);
          glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set wireframe color
          break;
        case 2: // white with wireframes
          glDisable(GL_TEXTURE_2D);
          glEnable(GL_POLYGON_OFFSET_LINE);
          glPolygonOffset(0.0, -4);
          break;
        case 3: // only wireframes
          glDisable(GL_TEXTURE_2D);
          glPolygonMode(GL_FRONT, GL_LINE);
          glDisable(GL_POLYGON_OFFSET_LINE);
          //glColor4f(1.0f, 1.0f, 1.0f, 1.0f);  // sets wireframe color
          glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set wireframe color
          break;
      }

    // using display lists slows rendering twice
    // if gWireframe==1 or gWireframe==2
    if (gWireframe==1)
      {
        glEnable(GL_TEXTURE_2D);
        glPolygonMode(GL_FRONT, GL_FILL);

        DrawRoam(gCamera[0].mLocation, UpdateGLStateWF1);
      }
    else if (gWireframe==2)
      {
        glPolygonMode(GL_FRONT, GL_FILL);
        glColor4f(1.0f, 1.0f, 1.0f, 1.0f);  // set fill color

        DrawRoam(gCamera[0].mLocation, UpdateGLStateWF2);
      }
    else
      DrawRoam(gCamera[0].mLocation, 0);
  }

void InitDisplayRoamStandard()
  {
    fprintf(gOutFile, "Using Standard ROAM drawing method\n");
    printf("Using Standard ROAM drawing method\n");
    SelectFastDisplay();

    // initialize position tracking
    // 0000
    //gDistanceSoFar = -1024.;
    gDistanceSoFar = -8.;
    gLastLocation[0]=gCamera[0].mLocation[0];
    gLastLocation[1]=gCamera[0].mLocation[1];
    gLastLocation[2]=gCamera[0].mLocation[2];

    // initialize triangle pool
    InitTriPool();

    sSplitQueue.Init();
    sMergeQueue.Init();

    // initialize base mesh
    int numroots=gModel->GetNumBaseTriangles();
    int i;
    gRoot=new RoamTree[numroots];
    for (i=0; i<numroots; ++i)
      {
        int tempverts[3];
        for (int j=0; j<3; ++j)
          tempverts[j] = gModel->GetBaseTriangleVert(i, j);
        gRoot[i].InitializeVerts(tempverts, 
                      i<gModel->GetAdjacentBaseTriangle(i, Model::Base));
      }
    for (i=0; i<numroots; ++i)
      {
        RoamTree *nbrs[3];
        for (int j=0; j<3; ++j)
          {
            int tempnbr = gModel->GetAdjacentBaseTriangle(i, j);
            assert(tempnbr<numroots);
            assert(tempnbr>=-1);
            if (tempnbr>=0)
              nbrs[j]=gRoot+tempnbr;
            else
              nbrs[j]=0;
          }
        gRoot[i].InitializeNeighbors(nbrs[0], nbrs[1], nbrs[2]);
      }

    gDisplay->Initialize(MaxVertsPerBlock, MaxIndicesPerBlock);
  }

static void RecursivelyDeleteTree(RoamTri *todel)
  {
    if (! todel[0].IsLeaf())
      {
        assert(todel[0].GetChildren());
        RecursivelyDeleteTree(todel[0].GetChildren());
      }
    if (! todel[1].IsLeaf() && ! todel[1].IsDummyBranch())  // todel[1] may be a dummy branch
      {
        RecursivelyDeleteTree(todel[1].GetChildren());
      }
   
    DeallocateTwoTris(todel);
  }

void CloseDisplayRoamStandard()
  {
    sSplitQueue.Clear();
    sMergeQueue.Clear();

    // release triangle pool
    // BOGUS: when rewrite tri pool, won't need the following:
    for (int i=0; i<gModel->GetNumBaseTriangles(); ++i)
      RecursivelyDeleteTree(gRoot[i].GetRootTriangle());
    CloseTriPool();

    gDisplay->Done();

    // free base mesh
    delete[] gRoot;
  }

void PrintStatsRoamStandard()
  {
    fprintf(gOutFile, "%.2f merge tests, %.2f merges, %.2f split tests, %.2f splits, %.2f force splits per frame\n", 
            ((double)gMergeTests)/gNumFrames, 
            ((double)gMerges)/gNumFrames, 
            ((double)gSplitTests)/gNumFrames, 
            ((double)gSplits)/gNumFrames, 
            ((double)gForceSplits)/gNumFrames);
#if 0
    fprintf(gOutFile, "%.2f fans/frame, %.3f tris/fan\n",
            ((double)gOutputFans)/gNumFrames,
            ((double)gNumOutputTris)/gOutputFans);
#endif
  }

void ResetStatsRoamStandard()
  {
    gMergeTests=0;
    gMerges=0;
    gSplitTests=0;
    gSplits=0;
    gForceSplits=0;
#if 0
    gOutputFans=0;
#endif
  }

// End Roam Standard
// -----------------------------------------------------------------------------------------------------

